# coding=utf-8
# Copyright (C) Alex Urban (2019)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Batch-mode utilities for omega scans
"""

import os
import subprocess
import sys

from getpass import getuser
from pycondor import (Dagman, Job)

from .. import (cli, condor)

# authorship credits
__author__ = 'Alex Urban <alexander.urban@ligo.org>'
__credits__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

# set CLI docstring
CLI_DOCSTRING = """Batch-generate a series of Omega scans

GPS times can be given individually on the command-line, one after the other,
or can be bundled into one file formatted where the first column contains
the GPS times (other columns are ignored).

The output of this script is a condor workflow in the form of a DAG file,
with associated condor submit (`.sub`) file in the output directory.
Submitting the workflow to Condor will result in the scans being processed
in parallel.
"""

# set default accounting information
ACCOUNTING_GROUP = os.getenv(
    '_CONDOR_ACCOUNTING_GROUP',
    'ligo.dev.{epoch}.detchar.user_req.omegascan',
)
ACCOUNTING_GROUP_USER = os.getenv(
    '_CONDOR_ACCOUNTING_USER',
    getuser(),
)

# set program name
PROG = ('python -m gwdetchar.omega.batch' if sys.argv[0].endswith('.py')
        else os.path.basename(sys.argv[0]))


# -- utilities ----------------------------------------------------------------

def _monitor_dag_workflow(dagman):
    """Monitor a batch of Omega scans through the Condor pool
    """
    print("Monitoring progress of {0.submit_file}".format(dagman))
    try:
        subprocess.check_call(
            ["pycondor", "monitor", dagman.submit_file],
        )
    except KeyboardInterrupt:
        pass


def _parse_analysis_times(tlist):
    """Parse an iterable of GPS time data for a batch of Omega scans
    """
    if len(tlist) == 1:
        try:  # try converting to GPS
            tlist = [float(tlist[0])]
        except (TypeError, ValueError):  # otherwise read as file
            import numpy
            tlist = numpy.loadtxt(tlist[0], dtype=float, ndmin=1)
    else:
        tlist = list(map(float, tlist))
    return tlist


def get_command_line_flags(ifo, fscale='log', colormap='viridis', nproc=8,
                           far=3.171e-8, config_file=None,
                           disable_correlation=False, disable_checkpoint=False,
                           ignore_state_flags=False):
    """Get a list of optional command-line arguments to `gwdetchar-omega`
    """
    flags = [
        "--ifo", ifo,
        "--frequency-scaling", fscale,
        "--colormap", colormap,
        "--nproc", str(nproc),
        "--far-threshold", str(far),
    ]
    if config_file is not None:
        flags.extend(("--config-file", os.path.abspath(config_file)))
    if disable_correlation:
        flags.append("--disable-correlation")
    if disable_checkpoint:
        flags.append("--disable-checkpoint")
    if ignore_state_flags:
        flags.append("--ignore-state-flags")
    return flags


def get_condor_arguments(accounting_group=ACCOUNTING_GROUP,
                         accounting_group_user=ACCOUNTING_GROUP_USER,
                         timeout=0, extra_commands=[], gps=0):
    """Get a list of arguments for Condor processing
    """
    # get reference epoch
    if '{epoch}' in accounting_group:
        epoch = condor.accounting_epoch(gps)
        accounting_group = accounting_group.format(epoch=epoch.lower())
    # validate accounting tag
    condor.is_valid(accounting_group)
    # determine condor arguments
    condorcmds = [
        "accounting_group = {}".format(accounting_group),
        "accounting_group_user = {}".format(accounting_group_user),
    ]
    if timeout:
        condorcmds.append("periodic_remove = {}".format(
            'CurrentTime-EnteredCurrentStatus > {}'.format(
                3600 * timeout),
        ))
    condorcmds.extend(extra_commands)
    return condorcmds


def generate_dag(times, flags=[], tag='gwdetchar-omega-batch',
                 submit=False, outdir=os.getcwd(), universe='vanilla',
                 condor_commands=get_condor_arguments()):
    """Construct a Directed Acyclic Graph (DAG) for a batch of omega scans

    Parameters
    ----------
    times : `list` of `float`
        list of GPS times to scan

    flags : `list` of `str`, optional
        a list of command-line flags to run for each job, defaults to an
        empty list

    tag : `str`, optional
        a helpful string to use to name the DAG,
        default: `'gwdetchar-omega-batch'`

    submit : `bool`, optional
        submit the DAG to condor, default: `False`

    outdir : `str`, optional
        the output directory in which to store files, will result in
        sub-directories called `'condor'` and `'log'`, default: `os.getcwd`

    universe : `str`, optional
        condor universe to run in, default: `'vanilla'`

    condor_commands : `list` of `str`, optional
        list of condor settings to process with, defaults to the output of
        `get_condor_arguments`

    Returns
    -------
    dagman : `~pycondor.Dagman`
        the fully built DAG object
    """
    logdir = os.path.join(outdir, 'logs')
    subdir = os.path.join(outdir, 'condor')
    # create DAG and jobs
    dagman = Dagman(name=tag, submit=subdir)
    job = Job(
        dag=dagman,
        name='gwdetchar-omega',
        executable=sys.executable,
        universe=universe,
        submit=subdir,
        error=logdir,
        output=logdir,
        getenv=True,
        request_memory=4096 if universe != "local" else None,
        extra_lines=condor_commands,
    )
    # make a node in the workflow for each event time
    for t in times:
        cmd = " ".join(["-m", "gwdetchar.omega", str(t)] + [
            "--output-directory", os.path.join(outdir, str(t))] + flags)
        job.add_arg(cmd, name=str(t).replace(".", "_"))
    # write and submit the DAG
    dagman.build(fancyname=False)
    print("Workflow generated for {} times".format(len(times)))
    if submit:
        dagman.submit_dag(submit_options="-force")
        print(
            "Submitted to condor, check status via:\n\n"
            "$ condor_q {}".format(getuser())
        )
    else:
        print(
            "Submit to condor via:\n\n"
            "$ condor_submit_dag {0.submit_file}".format(dagman),
        )
    return dagman


# -- parse command-line -------------------------------------------------------

def create_parser():
    """Create a command-line parser for this entry point
    """
    # initialize argument parser
    parser = cli.create_parser(
        prog=PROG,
        description=CLI_DOCSTRING,
    )
    pargs = parser.add_argument_group('Omega scan options')
    cargs = parser.add_argument_group('Condor options')

    # required arguments
    parser.add_argument(
        'gpstime',
        nargs='+',
        help='GPS time(s) to scan, or path to a file'
             'containing a single column of such times',
    )
    cli.add_ifo_option(
        parser,
    )
    parser.add_argument(
        '-o',
        '--output-dir',
        default=os.getcwd(),
        help='output directory for all scans, default: %(default)s',
    )

    # optional omega scan arguments
    pargs.add_argument(
        '-f',
        '--config-file',
        help='path to configuration file to use, default: '
             'choose based on observatory, epoch, and pipeline',
    )
    pargs.add_argument(
        '-d',
        '--disable-correlation',
        action='store_true',
        default=False,
        help='disable cross-correlation of aux '
             'channels, default: False',
    )
    pargs.add_argument(
        '-D',
        '--disable-checkpoint',
        action='store_true',
        default=False,
        help='disable checkpointing from previous '
             'runs, default: False',
    )
    pargs.add_argument(
        '-s',
        '--ignore-state-flags',
        action='store_true',
        default=False,
        help='ignore state flag definitions in '
             'the configuration, default: False',
    )
    pargs.add_argument(
        '-t',
        '--far-threshold',
        type=float,
        default=3.171e-8,
        help='white noise false alarm rate threshold (Hz) for '
             'processing channels, default: %(default)s',
    )
    pargs.add_argument(
        '-y',
        '--frequency-scaling',
        default='log',
        help='scaling of all frequency axes, default: %(default)s',
    )
    pargs.add_argument(
        '-c',
        '--colormap',
        default='viridis',
        help='name of colormap to use, default: %(default)s',
    )
    cli.add_nproc_option(
        pargs,
    )

    # optional condor arguments
    cargs.add_argument(
        '-u',
        '--universe',
        default='vanilla',
        type=str,
        help='universe for condor processing',
    )
    cargs.add_argument(
        '--submit',
        action='store_true',
        default=False,
        help='submit DAG directly to condor queue',
    )
    cargs.add_argument(
        '--monitor',
        action='store_true',
        default=False,
        help='monitor the DAG progress after submission; '
             'only used with --submit',
    )
    cargs.add_argument(
        '--condor-accounting-group',
        default=ACCOUNTING_GROUP,
        help='accounting_group for condor submission on the LIGO '
             'Data Grid, include \'{epoch}\' (with curly brackets) '
             'to auto-substitute the appropriate epoch based on '
             'the GPS times',
    )
    cargs.add_argument(
        '--condor-accounting-group-user',
        default=ACCOUNTING_GROUP_USER,
        help='accounting_group_user for condor submission on the '
             'LIGO Data Grid',
    )
    cargs.add_argument(
        '--condor-timeout',
        type=float,
        default=None,
        metavar='T',
        help='configure condor to terminate jobs after T hours '
             'to prevent idling, default: %(default)s',
    )
    cargs.add_argument(
        '--condor-command',
        action='append',
        default=[],
        help='Extra condor submit commands to add to '
             'gw_summary submit file. Can be given '
             'multiple times in the form "key=value"',
    )

    # return the argument parser
    return parser


# -- main code block ----------------------------------------------------------

def main(args=None):
    """Run the command-line Omega scan tool in batch mode
    """
    parser = create_parser()
    args = parser.parse_args(args=args)

    # set up output directory
    outdir = os.path.abspath(
        os.path.expanduser(args.output_dir))

    # parse analysis times
    times = _parse_analysis_times(
        getattr(args, 'gpstime'))

    # get condor arguments
    condorcmds = get_condor_arguments(
        accounting_group=args.condor_accounting_group,
        accounting_group_user=args.condor_accounting_group_user,
        timeout=args.condor_timeout,
        extra_commands=args.condor_command,
        gps=max(times),
    )

    # get command-line flags
    flags = get_command_line_flags(
        args.ifo,
        fscale=args.frequency_scaling,
        colormap=args.colormap,
        nproc=args.nproc,
        far=args.far_threshold,
        config_file=args.config_file,
        disable_correlation=args.disable_correlation,
        disable_checkpoint=args.disable_checkpoint,
        ignore_state_flags=args.ignore_state_flags,
    )

    # -- generate workflow ------------

    # write and submit the DAG
    dagman = generate_dag(
        times,
        flags=flags,
        tag="gwdetchar-omega-batch",
        submit=args.submit,
        outdir=outdir,
        universe=args.universe,
        condor_commands=condorcmds,
    )

    # monitor DAG progress
    if (args.submit and args.monitor):
        _monitor_dag_workflow(dagman)


# -- run from command-line ----------------------------------------------------

if __name__ == "__main__":
    main()



# coding=utf-8
# Copyright (C) LIGO Scientific Collaboration (2015-)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Produce an Omega scan for a list of channels around a given GPS time

This utility can be used to process an arbitrary list of detector channels
with minimal effort in finding data. The input should be an INI-formatted
configuration file that lists processing options and channels in contextual
blocks. For more information, see gwdetchar.omega.config.
"""

import numpy
import os
import sys
import warnings

from gwpy.table import Table
from gwpy.time import to_gps

from .. import (cli, omega)
from ..io.datafind import (check_flag, get_data)
from . import (config, html)

from matplotlib import use
use('Agg')

# backend-dependent import
from . import plot  # noqa: E402

# authorship credits
__author__ = 'Alex Urban <alexander.urban@ligo.org>'
__credits__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

# set up logger
PROG = ('python -m gwdetchar.omega' if sys.argv[0].endswith('.py')
        else os.path.basename(sys.argv[0]))
LOGGER = cli.logger(name=PROG.split('python -m ').pop())


# -- utilities ----------------------------------------------------------------

def _finalize_html(analyzed, ifo, gps, htmlv):
    """Write the final HTML data product for this analysis
    """
    htmlv['refresh'] = False  # turn off auto-refresh
    if analyzed:
        html.write_qscan_page(ifo, gps, analyzed, **htmlv)
        return
    reason = ('No significant channels found '
              'during active analysis segments')
    html.write_null_page(ifo, gps, reason, context=ifo.lower())


def _init_analyzed_channels():
    """Initialize a running, ordered record of analyzed channels
    """
    if sys.version_info >= (3, 7):  # python 3.7+
        return {}
    from collections import OrderedDict
    return OrderedDict()


def _load_channel_record(summary, use_checkpoint=True, correlate=True):
    """Load a previous Omega scan from its last saved checkpoint
    """
    if not (use_checkpoint and os.path.exists(summary)):
        return ([], [])
    LOGGER.debug('Checkpointing from {}'.format(
        os.path.abspath(summary)))
    record = Table.read(summary)
    if correlate and ('Standard Deviation' not in record.colnames):
        raise KeyError(
            'Cross-correlation is not available from this record, '
            'consider running without correlation or starting from '
            'scratch with --disable-checkpoint')
    completed = list(record['Channel'])
    return (record, completed)


def _load_channel_from_checkpoint(block_name, channel, analyzed,
                                  completed, record, correlated):
    """Load channel data from a previous analysis
    """
    LOGGER.info(' -- Checkpointing {} from a previous '
                'run'.format(channel.name))
    cindex = completed.index(channel.name)
    channel.load_loudest_tile_features(
        record[cindex], correlated=correlated)
    return html.update_toc(analyzed, channel, name=block_name)


def _parse_configuration(inifiles, ifo=None, gps=None):
    """Parse configuration files for this Omega scan
    """
    # get path(s) to configuration files
    inifiles = [os.path.abspath(f) for f in (
        inifiles or
        config.get_default_configuration(ifo, gps)
    )]
    # parse configuration files
    LOGGER.debug('Parsing the following configuration files:')
    for fname in inifiles:
        LOGGER.debug(''.join([' -- ', fname]))
    return inifiles


def _scan_channel(channel, data, analyzed, gps, block, fthresh,
                  logf, fscale, colormap, block_name, correlate=None):
    """Perform Omega scan on an individual data channel
    """
    try:  # scan the channel
        LOGGER.info(' -- Scanning channel {}'.format(channel.name))
        series = omega.scan(
            gps, channel, data.astype('float64'), block.fftlength,
            resample=block.resample, fthresh=fthresh, search=block.search,
            logf=logf)
    except (ValueError, KeyError) as exc:
        warnings.warn("Skipping {}: [{}] {}".format(
            channel.name, type(exc), str(exc)), UserWarning)
        return analyzed
    # if channel is insignificant, skip it
    if series is None:
        LOGGER.warning(
            ' -- Channel not significant at white noise false alarm '
            'rate {} Hz'.format(fthresh))
        return analyzed
    # plot Omega scan products
    LOGGER.info(
        ' -- Plotting Omega scan products for {}'.format(channel.name))
    plot.write_qscan_plots(gps, channel, series, fscale=fscale,
                           colormap=colormap)
    # handle cross-correlation
    if correlate is not None:
        LOGGER.info(' -- Cross-correlating {}'.format(channel.name))
        correlation = omega.cross_correlate(series[2], correlate)
        channel.save_loudest_tile_features(
            series[3], correlation, gps=gps, dt=block.dt)
    else:
        channel.save_loudest_tile_features(series[3])
    # update the record of analyzed channels
    return html.update_toc(analyzed, channel, block_name)


# -- parse command-line -------------------------------------------------------

def create_parser():
    """Create a command-line parser for this entry point
    """
    # initialize argument parser
    parser = cli.create_parser(
        prog=PROG,
        description=__doc__,
    )

    # required argument
    parser.add_argument(
        'gpstime',
        type=to_gps,
        help='GPS time or datestring to scan',
    )

    # optional arguments
    cli.add_ifo_option(
        parser,
        required=False,
    )
    parser.add_argument(
        '-o',
        '--output-directory',
        help='output directory for the Omega scan, '
             'default: ~/public_html/wdq/{IFO}_{gpstime}',
    )
    parser.add_argument(
        '-f',
        '--config-file',
        action='append',
        default=None,
        help='path to configuration file to use, can be given '
             'multiple times (files read in order), default: '
             'choose a standard one based on IFO and GPS time',
    )
    parser.add_argument(
        '-d',
        '--disable-correlation',
        action='store_true',
        default=False,
        help='disable cross-correlation of aux '
             'channels, default: False',
    )
    parser.add_argument(
        '-D',
        '--disable-checkpoint',
        action='store_true',
        default=False,
        help='disable checkpointing from previous '
             'runs, default: False',
    )
    parser.add_argument(
        '-s',
        '--ignore-state-flags',
        action='store_true',
        default=False,
        help='ignore state flag definitions in '
             'the configuration, default: False',
    )
    parser.add_argument(
        '-t',
        '--far-threshold',
        type=float,
        default=3.171e-8,
        help='white noise false alarm rate threshold (Hz) for '
             'processing channels, default: %(default)s',
    )
    parser.add_argument(
        '-y',
        '--frequency-scaling',
        default='log',
        help='scaling of all frequency axes, default: %(default)s',
    )
    parser.add_argument(
        '-c',
        '--colormap',
        default='viridis',
        help='name of colormap to use, default: %(default)s',
    )
    cli.add_nproc_option(
        parser,
    )

    # return the argument parser
    return parser


# -- main code block ----------------------------------------------------------

def main(args=None):
    """Run the Omega scan command-line tool
    """
    parser = create_parser()
    args = parser.parse_args(args=args)

    # enforce logging level
    LOGGER.setLevel('DEBUG')

    # get critical arguments
    ifo = args.ifo or 'Network'
    gps = numpy.around(float(args.gpstime), 2)
    LOGGER.info("{0} Omega Scan {1}".format(ifo, gps))

    # parse configuration files
    args.config_file = _parse_configuration(
        args.config_file, ifo=ifo, gps=gps)
    cp = config.OmegaConfigParser(ifo=ifo)
    cp.read(args.config_file)

    # parse primary channel
    if not args.disable_correlation:
        try:
            primary = config.OmegaChannelList(
                'primary', **dict(cp.items('primary')))
        except config.configparser.NoSectionError:
            LOGGER.warning('No primary configured, continuing '
                           'without cross-correlation')
            args.disable_correlation = True
    cp.remove_section('primary')

    # get contextual channel blocks
    blocks = cp.get_channel_blocks()

    # set up analyzed channel dict
    analyzed = _init_analyzed_channels()

    # prepare html variables
    htmlv = {
        'title': '{} Qscan | {}'.format(ifo, gps),
        'config': args.config_file,
        'prog': PROG,
        'refresh': True,
    }

    # set output directory
    indir = os.getcwd()
    outdir = os.path.abspath(
        args.output_directory or
        os.path.expanduser(
            '~/public_html/wdq/{ifo}_{gps}'.format(ifo=ifo, gps=gps),
        ))
    if not os.path.isdir(outdir):
        os.makedirs(outdir)
    os.chdir(outdir)
    LOGGER.debug('Output directory created as {}'.format(outdir))

    # make subdirectories
    for d in ['plots', 'about', 'data']:
        if not os.path.isdir(d):
            os.makedirs(d)

    # load checkpoints, if requested
    summary = os.path.join('data', 'summary.csv')
    (record, completed) = _load_channel_record(
        summary,
        use_checkpoint=(not args.disable_checkpoint),
        correlate=(not args.disable_correlation),
    )

    # set up html output
    LOGGER.debug('Setting up HTML at {}'.format(
        os.path.join(outdir, 'index.html')))
    html.write_qscan_page(ifo, gps, analyzed, **htmlv)

    # -- compute Q-scan ---------------

    # launch Omega scans
    LOGGER.info('Launching Omega scans')

    # construct a matched-filter from primary channel
    if not args.disable_correlation:
        LOGGER.debug('Processing primary channel')
        duration = primary.duration
        fftlength = primary.fftlength
        # process `duration` seconds of data centered on gps
        name = primary.channel.name
        start = gps - duration/2. - 1
        end = gps + duration/2. + 1
        correlate = get_data(
            name, start, end, frametype=primary.frametype,
            source=primary.source, nproc=args.nproc,
            verbose='Reading primary:'.rjust(30))
        correlate = omega.primary(
            gps, primary.length, correlate, fftlength,
            resample=primary.resample, f_low=primary.flow)
        plot.timeseries_plot(correlate, gps, primary.length, name,
                             'plots/primary.png', ylabel='Whitened Amplitude')
        # prepare HTML output
        htmlv['correlated'] = True
        htmlv['primary'] = name
    else:
        correlate = None

    # range over channel blocks
    for block in blocks.values():
        LOGGER.debug('Processing block {}'.format(block.key))
        chans = [c.name for c in block.channels]
        # get configuration
        duration = block.duration
        fftlength = block.fftlength
        # check that analysis flag is active for all of `duration`
        if block.flag and (not args.ignore_state_flags):
            LOGGER.info(' -- Querying state flag {}'.format(block.flag))
            if not check_flag(block.flag, gps, duration, pad=1):
                LOGGER.info(
                    ' -- {} not active, skipping block'.format(block.flag))
                continue
        # read `duration` seconds of data
        if not (set(chans) <= set(completed)):
            start = gps - duration/2. - 1
            end = gps + duration/2. + 1
            data = get_data(
                chans, start, end, frametype=block.frametype,
                source=block.source, nproc=args.nproc,
                verbose='Reading block:'.rjust(30))

        # process individual channels
        for channel in block.channels:
            if channel.name in completed:  # load checkpoint
                analyzed = _load_channel_from_checkpoint(
                    blocks[channel.section].name, channel, analyzed,
                    completed, record, (correlate is not None))
                htmlv['toc'] = analyzed
                html.write_qscan_page(ifo, gps, analyzed, **htmlv)
                continue

            analyzed = _scan_channel(
                channel, data[channel.name], analyzed, gps, block,
                args.far_threshold, (args.frequency_scaling == 'log'),
                args.frequency_scaling, args.colormap,
                blocks[channel.section].name, correlate)
            htmlv['toc'] = analyzed
            html.write_qscan_page(ifo, gps, analyzed, **htmlv)

    # -- finalize HTML ----------------

    # write HTML page and finish
    LOGGER.debug('Finalizing HTML at {}'.format(
        os.path.join(outdir, 'index.html')))
    _finalize_html(analyzed, ifo, gps, htmlv)
    LOGGER.info("-- index.html written, all done --")

    # return to original directory
    os.chdir(indir)


# -- run code -----------------------------------------------------------------

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright (C) Joshua Smith (2016-)
#
# This file is part of the GW DetChar python package.
#
# gwdetchar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# gwdetchar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gwdetchar.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwdetchar.daq`
"""

import numpy
import pytest

from numpy.testing import assert_array_equal
from unittest import mock

from gwpy.timeseries import TimeSeries
from gwpy.segments import (Segment, SegmentList)
from gwpy.testing.utils import assert_segmentlist_equal

from .. import daq

OVERFLOW_SERIES = TimeSeries([0, 0, 0, 1, 1, 0, 0, 1, 0, 1], dx=.5)
CUMULATIVE_SERIES = TimeSeries([0, 0, 0, 1, 2, 2, 2, 3, 3, 4], dx=.5)
OVERFLOW_TIMES = numpy.array([1.5, 3.5, 4.5])
OVERFLOW_SEGMENTS = SegmentList([
    Segment(1.5, 2.5),
    Segment(3.5, 4),
    Segment(4.5, 5),
])
CROSSING_TIMES = numpy.array([1.5, 2.5, 3.5, 4, 4.5])

CHANNELS = [
    'X1:TEST-CHANNEL_1',
    'X1:FEC-1_ADC_OVERFLOW_ACC_0_1',
    'X1:FEC-1_ADC_OVERFLOW_ACC_0_2',
    'X1:FEC-1_ADC_OVERFLOW_ACC_0_3',
    'X1:FEC-1_ADC_OVERFLOW_ACC_0_4',
    'X1:FEC-1_DAC_OVERFLOW_0_1',
    'X1:FEC-1_DAC_OVERFLOW_0_2',
    'X1:FEC-2_DAC_OVERFLOW_0_1',
    'X1:FEC-2_DAC_OVERFLOW_0_2',
]


@pytest.mark.parametrize('cmltv, series', [
    (False, OVERFLOW_SERIES),
    (True, CUMULATIVE_SERIES),
])
def test_find_overflows_and_segments(cmltv, series):
    # find overflows
    times = daq.find_overflows(series, cumulative=cmltv)
    assert_array_equal(times, OVERFLOW_TIMES)

    # find segments
    segments = daq.find_overflow_segments(series, cumulative=cmltv)
    assert_segmentlist_equal(segments.active, OVERFLOW_SEGMENTS)


def test_ligo_accum_overflow_channel():
    assert daq.ligo_accum_overflow_channel(4, ifo='X1') == (
        'X1:FEC-4_ACCUM_OVERFLOW')


@mock.patch('gwdetchar.daq.find_urls')
@mock.patch('gwdetchar.daq.get_channel_names')
@mock.patch('gwdetchar.daq._ligo_model_overflow_channels_nds')
def test_ligo_model_overflow_channels(nds, get_names, find_frames):
    get_names.return_value = CHANNELS

    names = daq.ligo_model_overflow_channels(1, ifo='X1', accum=True)
    assert names == CHANNELS[1:5]

    names = daq.ligo_model_overflow_channels(1, ifo='X1', accum=False)
    assert names == CHANNELS[5:7]

    nds.return_value = CHANNELS
    names = daq.ligo_model_overflow_channels(1, ifo='X1', accum=False)
    assert names == CHANNELS[5:7]

    find_frames.return_value = []
    with pytest.raises(IndexError) as exc:
        daq.ligo_model_overflow_channels(1, ifo='X1')
    assert str(exc.value).startswith('No X-X1_R frames found')


def test_find_crossings():
    times = daq.find_crossings(OVERFLOW_SERIES, 0.1)
    assert_array_equal(times, CROSSING_TIMES)

    times0 = daq.find_crossings(OVERFLOW_SERIES, 0)
    assert_array_equal(times0, [])



# coding=utf-8
# Copyright (C) Alex Urban (2018-)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Plotting routines for omega scans
"""

from matplotlib import (cm, rcParams)

from gwpy.plot.colors import GW_OBSERVATORY_COLORS

from ..plot import texify

__author__ = 'Alex Urban <alexander.urban@ligo.org>'
__credits__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

rcParams.update({
    'savefig.transparent': False,
})


# -- internal formatting tools ------------------------------------------------

def _format_time_axis(ax, gps, span):
    """Format the time axis of an omega scan plot

    Parameters
    ----------
    ax : `~matplotlib.axis.Axis`
        the `Axis` object to format

    gps : `float`
        reference GPS time (in seconds) to serve as the origin

    span : `float`
        total duration (in seconds) of the time axis
    """
    # set time axis units
    ax.set_xscale('seconds', epoch=gps)
    ax.set_xlim(gps-span/2, gps+span/2)
    ax.grid(True, axis='x', which='major')


def _format_frequency_axis(ax, yscale='log'):
    """Format the frequency axis of an omega scan plot

    Parameters
    ----------
    ax : `~matplotlib.axis.Axis`
        the `Axis` object to format
    """
    ax.grid(True, axis='y', which='both')
    ax.set_yscale(yscale)
    ax.set_ylabel('Frequency [Hz]')


def _format_color_axis(ax, colormap='viridis', clim=None, norm='linear'):
    """Format the color axis of an omega scan spectral plot

    Parameters
    ----------
    ax : `~matplotlib.axis.Axis`
        the `Axis` object to format

    colormap : `str`
        matplotlib colormap to use, default: viridis

    clim : `tuple` or `None`
        limits of the color axis, default: autoscale with log scaling

    norm : `str`
        scaling of the color axis, only used if `clim` is given,
        default: linear
    """
    cmap = cm.get_cmap(colormap)
    ax.set_facecolor(cmap(0))
    # set colorbar format
    if clim is None:  # force a log colorbar with autoscaled limits
        ax.colorbar(cmap=colormap, norm='log', vmin=0.5,
                    label='Normalized energy')
    else:
        ax.colorbar(cmap=colormap, norm=norm, clim=clim,
                    label='Normalized energy')


# -- utilities ----------------------------------------------------------------

def timeseries_plot(data, gps, span, channel, output, ylabel=None,
                    figsize=(12, 6)):
    """Custom plot for a GWPy TimeSeries object

    Parameters
    ----------
    data : `~gwpy.timeseries.TimeSeries`
        the series to plot

    gps : `float`
        reference GPS time (in seconds) to serve as the origin

    span : `float`
        total duration (in seconds) of the time axis

    channel : `str`
        name of the channel corresponding to this data

    output : `str`
        name of the output file

    ylabel : `str` or `None`
        label for the y-axis

    figsize : `tuple`
        size (width x height) of the final figure, default: `(12, 6)`
    """
    # set color by IFO
    ifo = channel[:2]
    data = data.crop(gps-span/2, gps+span/2)
    plot = data.plot(color=GW_OBSERVATORY_COLORS[ifo], figsize=figsize)
    # set axis properties
    ax = plot.gca()
    _format_time_axis(ax, gps=gps, span=span)
    ax.set_yscale('linear')
    ax.set_ylabel(ylabel)
    # set title
    title = texify(channel)
    ax.set_title(title)
    # save plot and close
    plot.savefig(output)
    plot.close()


def spectral_plot(data, gps, span, channel, output, colormap='viridis',
                  clim=None, nx=1400, yscale='log', norm='linear',
                  figsize=(12, 6)):
    """Custom plot for a GWPy spectrogram or Q-gram

    Parameters
    ----------
    data : `~gwpy.timeseries.TimeSeries`
        the series to plot

    gps : `float`
        reference GPS time (in seconds) to serve as the origin

    span : `float`
        total duration (in seconds) of the time axis

    channel : `str`
        name of the channel corresponding to this data

    output : `str`
        name of the output file

    colormap : `str`
        matplotlib colormap to use, default: viridis

    clim : `tuple` or `None`
        limits of the color axis, default: autoscale with log scaling

    yscale : `str`
        scaling of the frequency axis, default: log

    norm : `str`
        scaling of the color axis, only used if `clim` is given,
        default: linear

    nx : `int`
        number of points along the time axis, default: 500

    figsize : `tuple`
        size (width x height) of the final figure, default: `(12, 6)`
    """
    from gwpy.spectrogram import Spectrogram
    # construct plot
    if isinstance(data, Spectrogram):
        # plot interpolated spectrogram
        Q = data.q
        data = data.crop(gps-span/2, gps+span/2)
        nslice = max(1, int(data.shape[0] / nx))
        plot = data[::nslice].pcolormesh(figsize=figsize)
    else:
        # plot eventgram
        Q = data.meta['q']
        plot = data.tile('time', 'frequency', 'duration', 'bandwidth',
                         color='energy', figsize=figsize, antialiased=True)
    # set axis properties
    ax = plot.gca()
    _format_time_axis(ax, gps=gps, span=span)
    _format_frequency_axis(ax, yscale=yscale)
    # set colorbar properties
    _format_color_axis(ax, colormap=colormap, clim=clim, norm=norm)
    # set title
    title = '{0} with $Q$ of {1:.1f}'.format(texify(channel), Q)
    ax.set_title(title)
    # save plot and close
    plot.savefig(output)
    plot.close()


def write_qscan_plots(gps, channel, series, fscale='log', colormap='viridis'):
    """Custom plot utility for a full omega scan

    Parameters
    ----------
    gps : `float`
        reference GPS time (in seconds) to serve as the origin

    channel : `OmegaChannel`
        channel corresponding to these data

    series : `tuple`
        a collection of `TimeSeries`, `Spectrogram`, and `QGram` objects

    fscale : `str`
        scaling of the frequency axis, default: log

    colormap : `str`, optional
        matplotlib colormap to use, default: viridis
    """
    # unpack series objects
    xoft, hpxoft, wxoft, qgram, rqgram, qspec, rqspec = series
    # range over plot types
    fnames = channel.plots
    for span, png1, png2, png3, png4, png5, png6, png7, png8, png9 in zip(
        channel.pranges, fnames['qscan_whitened'],
        fnames['qscan_autoscaled'], fnames['qscan_highpassed'],
        fnames['timeseries_raw'], fnames['timeseries_highpassed'],
        fnames['timeseries_whitened'], fnames['eventgram_highpassed'],
        fnames['eventgram_whitened'], fnames['eventgram_autoscaled']
    ):
        # plot whitened qscan
        spectral_plot(
            qspec, gps, span, channel.name, str(png1), clim=(0, 25),
            yscale=fscale, colormap=colormap)
        # plot autoscaled, whitened qscan
        spectral_plot(qspec, gps, span, channel.name, str(png2),
                      yscale=fscale, colormap=colormap)
        # plot raw qscan
        spectral_plot(
            rqspec, gps, span, channel.name, str(png3), clim=(0, 25),
            yscale=fscale, colormap=colormap)
        # plot raw timeseries
        timeseries_plot(xoft, gps, span, channel.name, str(png4),
                        ylabel='Amplitude')
        # plot highpassed timeseries
        timeseries_plot(hpxoft, gps, span, channel.name, str(png5),
                        ylabel='Highpassed Amplitude')
        # plot whitened timeseries
        timeseries_plot(wxoft, gps, span, channel.name, str(png6),
                        ylabel='Whitened Amplitude')
        # plot raw eventgram
        rtable = rqgram.table(snrthresh=channel.snrthresh)
        spectral_plot(
            rtable, gps, span, channel.name, str(png7), clim=(0, 25),
            yscale=fscale, colormap=colormap)
        # plot whitened eventgram
        table = qgram.table(snrthresh=channel.snrthresh)
        spectral_plot(
            table, gps, span, channel.name, str(png8), clim=(0, 25),
            yscale=fscale, colormap=colormap)
        # plot autoscaled whitened eventgram
        spectral_plot(table, gps, span, channel.name, str(png9),
                      yscale=fscale, colormap=colormap)
    return



# -*- coding: utf-8 -*-
# Copyright (C) Alex Urban (2019)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gwdetchar.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for `gwdetchar.omega.core`
"""

import numpy
from numpy import testing as nptest
from scipy import signal

from gwpy.timeseries import TimeSeries
from gwpy.signal.qtransform import QGram
from gwpy.spectrogram import Spectrogram

from .. import (core, config)

__author__ = 'Alex Urban <alexander.urban@ligo.org>'


# global test objects

FFTLENGTH = 8

NOISE = TimeSeries(
    numpy.random.normal(loc=1, scale=.5, size=16384 * 68),
    sample_rate=16384, epoch=-34).zpk([], [0], 1)
GLITCH = TimeSeries(
    signal.gausspulse(numpy.arange(-1, 1, 1./16384), bw=100),
    sample_rate=16384, epoch=-1) * 1e-4
INPUT = NOISE.inject(GLITCH)

CONFIGURATION = {
    'q-range': '3.3166,150',
    'frequency-range': '4.0,Inf',
    'plot-time-durations': '4',
    'always-plot': 'True',
}
CHANNEL = config.OmegaChannel(
    channelname='X1:TEST-STRAIN', section='test', **CONFIGURATION)


# -- unit tests ---------------------------------------------------------------

def test_highpass():
    # high-pass filter the input
    hp = core.highpass(INPUT, f_low=4)
    assert INPUT.is_compatible(hp)
    assert hp.size == INPUT.size
    assert abs(INPUT.value.mean()) > 10
    nptest.assert_almost_equal(hp.value.mean(), 0, decimal=5)


def test_whiten():
    # whiten the input
    whitened = core.whiten(INPUT, fftlength=FFTLENGTH)
    assert INPUT.is_compatible(whitened)
    assert whitened.size == INPUT.size

    # test the whitened timeseries with the glitch cropped out
    ind = list(range(33 * 16384)) + list(range(35 * 16384, whitened.size))
    nptest.assert_almost_equal(whitened.value[ind].mean(), 0, decimal=4)
    nptest.assert_almost_equal(whitened.value[ind].std(), 1, decimal=1)

    # test that the peak occurs at the expected time
    tmax = whitened.times[whitened.argmax()]
    assert tmax.value == 0


def test_conditioner():
    # get conditioned data products
    wxoft, xoft = core.conditioner(INPUT, fftlength=FFTLENGTH, resample=4096)
    assert isinstance(wxoft, TimeSeries)
    assert isinstance(xoft, TimeSeries)
    assert xoft.sample_rate.value == 4096
    assert xoft.is_compatible(wxoft)


def test_conditioner_with_highpass():
    # get conditioned data with highpassing
    wxoft, hpxoft, xoft = core.conditioner(
        INPUT, fftlength=FFTLENGTH, resample=4096, f_low=4)
    assert isinstance(wxoft, TimeSeries)
    assert isinstance(hpxoft, TimeSeries)
    assert xoft.is_compatible(wxoft)
    assert xoft.is_compatible(hpxoft)
    nptest.assert_almost_equal(hpxoft.value.mean(), 0, decimal=5)
    nptest.assert_almost_equal(wxoft.value.mean(), 0, decimal=2)


def test_primary():
    # condition a data stream for use as a matched-filter
    length = 6
    mfilter = core.primary(0, length=length, hoft=INPUT, fftlength=FFTLENGTH)
    assert mfilter.duration.value == length
    assert mfilter.times[mfilter.argmax()].value == 0

    # test again, with a high-pass filter
    f_low = 4
    mfilterhp = core.primary(0, length=length, f_low=f_low, hoft=INPUT,
                             fftlength=FFTLENGTH)
    assert mfilterhp.duration.value == length
    assert mfilterhp.times[mfilterhp.argmax()].value == 0


def test_cross_correlate():
    # cross-correlate two data streams
    wxoft = core.whiten(INPUT, fftlength=FFTLENGTH)
    whoft = core.primary(0, length=6, hoft=INPUT, fftlength=FFTLENGTH)
    corr = core.cross_correlate(wxoft, whoft)
    assert wxoft.is_compatible(corr)
    tmax = corr.abs().times[corr.argmax()]
    assert tmax.value == 0

    # test with resampling
    corr2 = core.cross_correlate(wxoft.resample(2048), whoft)
    corr3 = core.cross_correlate(wxoft, whoft.resample(2048))
    assert corr2.sample_rate.value == 2048
    assert corr2.is_compatible(corr3)
    assert corr2.abs().times[corr2.argmax()].value == tmax.value
    assert corr3.abs().times[corr3.argmax()].value == tmax.value


def test_scan():
    # get omega scan products
    xoft, hpxoft, wxoft, qgram, rqgram, qspec, rqspec = core.scan(
        gps=0, channel=CHANNEL, xoft=INPUT, resample=2048, fftlength=FFTLENGTH)
    assert isinstance(xoft, TimeSeries)
    assert isinstance(hpxoft, TimeSeries)
    assert isinstance(wxoft, TimeSeries)
    assert isinstance(qgram, QGram)
    assert isinstance(rqgram, QGram)
    assert isinstance(qspec, Spectrogram)
    assert isinstance(rqspec, Spectrogram)

    # test features
    assert qspec.shape == (1400, 700)
    assert qspec.shape == rqspec.shape
    assert qspec.q == rqspec.q
    assert qspec.q == qgram.plane.q
    assert qgram.plane.q == rqgram.plane.q

    # test low false alarm rate
    CHANNEL.always_plot = False
    empty = core.scan(gps=0, channel=CHANNEL, xoft=NOISE, resample=2048,
                      fftlength=FFTLENGTH)
    assert empty is None



# -*- coding: utf-8 -*-
# Copyright (C) Joshua Smith (2016-)
#
# This file is part of the GW DetChar python package.
#
# gwdetchar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# gwdetchar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gwdetchar.  If not, see <http://www.gnu.org/licenses/>.

"""Test suite for `gwdetchar.cli`
"""

import logging
import argparse
try:
    from importlib import reload
except ImportError:  # python < 3
    reload = reload

import pytest

from .. import (cli, __version__ as gwdetchar_version, const as _const)


def test_logger():
    logger = cli.logger()
    assert isinstance(logger, logging.Logger)
    assert logger.name == 'gwdetchar.cli'
    assert logger.level == logging.DEBUG


@pytest.fixture
def const():
    yield _const
    reload(_const)


@pytest.fixture
def parser():
    return argparse.ArgumentParser()


def test_create_parser():
    parser = cli.create_parser(description=__doc__)
    assert isinstance(parser, argparse.ArgumentParser)
    assert parser.description == __doc__
    assert parser._actions[-1].version == gwdetchar_version


@pytest.mark.parametrize('inv, outv', [
    (None, gwdetchar_version),
    ('test', 'test'),
])
def test_add_version_option(parser, inv, outv):
    act = cli.add_version_option(parser, version=inv)
    assert act.version == outv
    with pytest.raises(SystemExit):
        parser.parse_args(['--version'])


@pytest.mark.parametrize('ifo', (_const.IFO, None))
def add_ifo_option(parser, const, ifo):
    const.IFO == ifo
    parser.add_ifo_option()
    args = parser.parse_args([])
    assert args.ifo == ifo
    if ifo is None:
        with pytest.raises(argparse.ArgumentError):
            parser.parse_args([])
    assert parser.parse_args(['--ifo', 'test']).ifo == 'test'


def test_add_gps_start_stop_arguments(parser):
    cli.add_gps_start_stop_arguments(parser)
    args = parser.parse_args(['Jan 1 2000', 'Jan 2 2000'])
    assert args.gpsstart == 630720013
    assert args.gpsend == 630806413


def test_add_gps_start_stop_options(parser):
    cli.add_gps_start_stop_options(parser)
    args = parser.parse_args([])
    assert args.gps_start_time is None
    assert args.gps_end_time is None
    args = parser.parse_args(['-s', 'Jan 1 2000',
                              '--gps-end-time', 'Jan 2 2000'])
    assert args.gps_start_time == 630720013
    assert args.gps_end_time == 630806413


def test_add_frame_type_name(parser):
    cli.add_frametype_option(parser)
    args = parser.parse_args(['--frametype', 'test'])
    assert args.frametype == 'test'


def test_add_nproc_option(parser):
    cli.add_nproc_option(parser)
    assert parser.parse_args([]).nproc == 8
    assert parser.parse_args(['-j', '2']).nproc == 2



# coding=utf-8
# Copyright (C) Duncan Macleod (2015)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Utilies for analysing sensor saturations
"""

import re
from itertools import zip_longest
from multiprocessing import (cpu_count, Pool)

import numpy
from astropy.units import Quantity

from gwpy.io.cache import file_segment
from gwpy.timeseries import StateTimeSeries

from ..io.datafind import get_data

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__credits__ = 'Dan Hoak <daniel.hoak@ligo.org>' \
              'Alex Urban <alexander.urban@ligo.org>'

DEFAULT_NPROC = min(8, cpu_count())

re_limit = re.compile(r'_LIMIT\Z')
re_limen = re.compile(r'_LIMEN\Z')
re_swstat = re.compile(r'_SWSTAT\Z')
re_software = re.compile(
    '(%s)' % '|'.join([re_limit.pattern, re_limen.pattern, re_swstat.pattern]))


# -- utilities ----------------------------------------------------------------

def _find_saturations(data):
    """Thin wrapper around find_saturations
    """
    out = find_saturations(
        data[0], data[1], precision=.99, segments=True)
    out.name = out.name[:-7]
    return out


def find_saturations(timeseries, limit=2**16, precision=1, segments=False):
    """Find the times of software saturations of the given `TimeSeries`

    Parameters
    ----------
    timeseries : `~gwpy.timeseries.TimeSeries`
        the input data to search

    limit : `float`, `~gwpy.timeseries.TimeSeries`
        the limit above which a saturation has occurred

    precision : `float` in range (0, 1]
        the precision of the check for saturation

    segments : `bool`, default: `False`
        return saturation segments, otherwise return times when
        saturations occur

    Returns
    -------
    times : `numpy.ndarray`
        the array of times when this timeseries started saturating, OR

    segments : `~gwpy.segments.DataQualityFlag`
        the flag containing segments during which this timeseries
        was actively saturating
    """
    if isinstance(limit, Quantity):
        limit = limit.value
    # check saturated at minimum and maximum
    limit = limit * precision
    saturated = timeseries.value <= -limit
    saturated |= timeseries.value >= limit
    if segments:
        saturation = saturated.view(StateTimeSeries)
        saturation.__metadata_finalize__(timeseries)
        flag = saturation.to_dqflag(
            description="Software saturation indicated by " + timeseries.name,
        )
        flag.isgood = False
        return flag
    else:
        return timeseries.times[1:].value[
            numpy.diff(saturated.astype(int)) > 0]


def grouper(iterable, n, fillvalue=None):
    """Separate an iterable into sub-sets of `n` elements
    """
    args = [iter(iterable)] * n
    return zip_longest(fillvalue=fillvalue, *args)


def find_limit_channels(channels, skip=None):
    """Find all 'LIMIT' channels that have a matching 'LIMEN' or 'SWSTAT'

    Parameters
    ----------
    channels : `list` of `str`
        the list of channel names to search

    Returns
    -------
    limits : `list` or `str`
        the list of channels whose name ends in '_LIMIT' for whom a matching
        channel ending in '_LIMEN' or '_SWSTAT' was found
    """
    # find relevant channels and sort them
    if skip:
        re_skip = re.compile('(%s)' % '|'.join(skip))
        useful = sorted(x for x in channels if re_software.search(x) and
                        not re_skip.search(x))
    else:
        useful = sorted(x for x in channels if re_software.search(x))

    # map limits to limen or swstat
    limits = sorted(x[:-6] for x in useful if re_limit.search(x))
    limens = sorted(x[:-6] for x in useful if re_limen.search(x)
                    and x[:-6] in limits)
    swstats = sorted(x[:-7] for x in useful if re_swstat.search(x)
                     and x[:-7] in limits)
    return (limens, swstats)


def is_saturated(channel, cache, start=None, end=None, indicator='LIMEN',
                 nproc=DEFAULT_NPROC):
    """Check whether a channel has saturated its software limit

    Parameters
    ----------
    channel : `str`, or `list` of `str`
        either a single channel name, or a list of channel names

    cache : `list`
        a `list` of file paths, the cache must be contiguous

    start : `~gwpy.time.LIGOTimeGPS`, `int`
        the GPS start time of the check

    end : `~gwpy.time.LIGOTimeGPS`, `int`
        the GPS end time of the check

    indicator : `str`
        the suffix of the indicator channel, either `'LIMEN'` or `'SWSTAT'`

    nproc : `int`
        the number of parallel processes to use for frame reading

    Returns
    -------
    saturated : `bool`, `None`, or `DataQualityFlag`, or `list` of the same
        one of the following given the conditions
        - `None` : if the channel doesn't have a software limit
        - `False` : if the channel didn't saturate
        - `~gwpy.segments.DataQualityFlag` : otherwise
        OR, a `list` of the above if a `list` of channels was given in the
        first place
    """
    if isinstance(channel, (list, tuple)):
        channels = channel
    else:
        channels = [channel]
    # parse prefix
    for i, c in enumerate(channels):
        if c.endswith('_LIMIT'):
            channels[i] = c[:-6]
    # check limit if set
    indicators = ['{}_{}'.format(c, indicator) for c in channels]
    gps = file_segment(cache[0])[0]
    data = get_data(indicators, gps, gps+1, source=cache, nproc=nproc)

    # check limits for returned channels
    if len(data) < len(channels):  # exclude nonexistent channels
        channels = [
            c for c in channels if '{}_{}'.format(c, indicator) in data]
        indicators = ['{}_{}'.format(c, indicator) for c in channels]
    if indicator.upper() == 'LIMEN':
        active = dict((c, data[indicators[i]].value[0]) for
                      i, c in enumerate(channels))
    elif indicator.upper() == 'SWSTAT':
        active = dict(
            (c, data[indicators[i]].astype('uint32').value[0] >> 13 & 1) for
            i, c in enumerate(channels))
    else:
        raise ValueError("Don't know how to determine if limit is set for "
                         "indicator %r" % indicator)
    # get output/limit data for all with active limits
    activechans = [c for c in channels if active[c]]
    datachans = ['%s_%s' % (c, s) for c in activechans for
                 s in ('LIMIT', 'OUTPUT')]
    data = get_data(datachans, start, end, source=cache, nproc=nproc)

    # find saturations of the limit for each channel
    dataiter = ((data['%s_OUTPUT' % c], data['%s_LIMIT' % c])
                for c in activechans)
    if nproc > 1:
        with Pool(processes=nproc) as pool:
            saturations = list(pool.map(_find_saturations, dataiter))
    else:
        saturations = list(map(_find_saturations, dataiter))

    # return many or one (based on input)
    if isinstance(channel, (list, tuple)):
        return saturations
    else:
        return saturations[0]



# coding=utf-8
# Copyright (C) Alex Urban (2019)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Methods and utilties for Lasso correlation
"""

__author__ = 'Alex Urban <alexander.urban@ligo.org>'

# -- imports ------------------------------------------------------------------

# import lasso utils
from .core import (
    find_outliers,
    remove_outliers,
    fit,
    find_alpha,
    remove_flat,
    remove_bad,
)



# -*- coding: utf-8 -*-
# Copyright (C) Joshua Smith (2016-)
#
# This file is part of the GW DetChar python package.
#
# gwdetchar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# gwdetchar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gwdetchar.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwdetchar.saturation.core`
"""

import pytest

import numpy
from numpy.testing import assert_array_equal

from unittest import mock

from gwpy.segments import (Segment, SegmentList, DataQualityFlag)
from gwpy.timeseries import (TimeSeries, TimeSeriesDict)
from gwpy.testing.utils import assert_segmentlist_equal

from .. import core

# global test objects

DATA = TimeSeries([1, 2, 3, 4, 5, 5, 5, 4, 5, 4], dx=.5, name='X1:TEST_OUTPUT')
SATURATIONS = numpy.array([2., 4.])
SEGMENTS = SegmentList([
    Segment(2., 3.5),
    Segment(4., 4.5),
])

CHANNELS = [
    'X1:TEST_LIMIT',
    'X1:TEST_LIMEN',
    'X1:TEST_SWSTAT',
]
TSDICT = TimeSeriesDict({
    'X1:TEST_LIMEN': TimeSeries(numpy.ones(10), dx=.5, name='X1:TEST_LIMEN'),
    'X1:TEST_OUTPUT': DATA,
    'X1:TEST_LIMIT': TimeSeries(5*numpy.ones(10), dx=.5, name='X1:TEST_LIMIT'),
})


# -- unit tests ---------------------------------------------------------------

def test_find_saturations():
    sats = core.find_saturations(DATA, limit=5., segments=False)
    assert_array_equal(sats, SATURATIONS)
    segs = core.find_saturations(DATA, limit=5.*DATA.unit, segments=True)
    assert_segmentlist_equal(segs.active, SEGMENTS)


def test_find_saturations_wrapper():
    segs = core._find_saturations((DATA, 5.))
    assert_segmentlist_equal(segs.active, SEGMENTS)
    assert segs.name == DATA.name[:-7]


def test_grouper():
    in_ = range(100)
    grouped = list(core.grouper(in_, 5))
    expected = [tuple(range(i, i+5)) for i in range(0, 100, 5)]
    assert_array_equal(grouped, expected)


def test_find_limit_channels():
    limens, swstats = core.find_limit_channels(CHANNELS)
    assert len(limens) == len(swstats)
    assert_array_equal(limens, ['X1:TEST'])
    assert_array_equal(swstats, ['X1:TEST'])

    limens2, swstats2 = core.find_limit_channels(CHANNELS, skip='TEST')
    assert not limens2
    assert not swstats2


@mock.patch('gwdetchar.io.datafind.remove_missing_channels')
@mock.patch('gwpy.timeseries.TimeSeriesDict.read')
def test_is_saturated(tsdfetch, remove):
    cache = [
        "X-TEST-0-1.gwf",
    ]

    tsdfetch.return_value = TSDICT
    remove.return_value = ['X1:TEST_LIMIT']

    saturated = core.is_saturated('X1:TEST', cache, start=0, end=8)
    assert isinstance(saturated, DataQualityFlag)
    assert_segmentlist_equal(saturated.active, SEGMENTS)

    saturated2 = core.is_saturated(
        ['X1:TEST_LIMIT'], cache, start=0, end=8)
    assert isinstance(saturated2, list)
    assert_segmentlist_equal(saturated2[0].active, saturated.active)

    with pytest.raises(ValueError) as exc:
        core.is_saturated(
            'X1:TEST', cache, start=0, end=8, indicator='blah')
    assert str(exc.value).startswith("Don't know how to determine")



# coding=utf-8
# Copyright (C) LIGO Scientific Collaboration (2015-)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Search for evidence of beam scattering based on optic velocity

This tool identifies broad time segments over which evidence for scattering is
strong. To compare projected fringes against spectrogram measurements for a
specific, narrow time range, please use the `simple` command-line module:

python -m gwdetchar.scattering.simple --help
"""

import numpy
import os.path
import random
import re
import sys
import warnings

from io import StringIO
from MarkupPy import markup

import gwtrigfind

from gwpy.segments import (
    DataQualityFlag,
    DataQualityDict,
    Segment,
    SegmentList,
)
from gwpy.table import EventTable
from gwpy.table.filters import in_segmentlist

from .. import cli
from ..const import DEFAULT_SEGMENT_SERVER
from ..io import html as htmlio
from ..io.datafind import get_data
from ..omega import batch

from . import (
    FREQUENCY_MULTIPLIERS,
    OPTIC_MOTION_CHANNELS,
    TRANSMON_CHANNELS,
    HIST_CAPTION,
    SCATTER_CAPTION,
    get_blrms,
    get_fringe_frequency,
    get_segments,
)

from matplotlib import (use, rcParams)
use('Agg')

# backend-dependent imports
from gwpy.plot import Plot  # noqa: E402
from ..plot import texify  # noqa: E402

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__credits__ = 'Siddharth Soni <siddharth.soni@ligo.org>' \
              'Alex Urban <alexander.urban@ligo.org>'

# set program name
PROG = ('python -m gwdetchar.scattering' if sys.argv[0].endswith('.py')
        else os.path.basename(sys.argv[0]))

# rcParams settings
rcParams.update({
    'axes.labelsize': 20,
    'figure.subplot.bottom': 0.17,
    'figure.subplot.left': 0.1,
    'figure.subplot.right': 0.9,
    'figure.subplot.top': 0.90,
    'grid.color': 'gray',
    'image.cmap': 'viridis',
    'svg.fonttype': 'none',
})


# -- parse command-line -------------------------------------------------------

def create_parser():
    """Create a command-line parser for this entry point
    """
    # initialize argument parser
    parser = cli.create_parser(
        prog=PROG,
        description=__doc__,
    )

    # required arguments
    cli.add_gps_start_stop_arguments(parser)
    cli.add_ifo_option(parser)

    # optional arguments
    parser.add_argument(
        '-a',
        '--state-flag',
        metavar='FLAG',
        default=None,
        help='restrict search to times when FLAG was active',
    )
    parser.add_argument(
        '-z',
        '--fmin',
        metavar='FMIN',
        type=float,
        default=10.,
        help='minimum frequency (Hz) of triggers, '
             'default: %(default)s',
    )
    parser.add_argument(
        '-t',
        '--frequency-threshold',
        type=float,
        default=15.,
        help='critical fringe frequency threshold (Hz), '
             'default: %(default)s',
    )
    parser.add_argument(
        '-x',
        '--multiplier-for-threshold',
        type=int,
        default=4,
        choices=FREQUENCY_MULTIPLIERS,
        help='fringe frequency multiplier to use when applying '
             '--frequency-threshold, default: %(default)s',
    )
    parser.add_argument(
        '-m',
        '--optic',
        action='append',
        help='optic to search for scattering signal, can be given '
             'multiple times, default: {}'.format(
                 list(OPTIC_MOTION_CHANNELS.keys())),
    )
    parser.add_argument(
        '-p',
        '--segment-padding',
        type=float,
        default=.05,
        help='time with which to pad scattering segments on '
             'either side, default: %(default)s',
    )
    parser.add_argument(
        '-s',
        '--segment-start-pad',
        type=float,
        default=30.0,
        help='amount of time to remove from the start of each '
             'analysis segment',
    )
    parser.add_argument(
        '-e',
        '--segment-end-pad',
        type=float,
        default=30.0,
        help='amount of time to remove from the end of each '
             'analysis segment',
    )
    parser.add_argument(
        '-L',
        '--bandpass-flow',
        type=float,
        default=4.0,
        help='lower-limit (Hz) of the passband for whitened '
             'BLRMS, default: %(default)s',
    )
    parser.add_argument(
        '-H',
        '--bandpass-fhigh',
        type=float,
        default=10.0,
        help='upper-limit (Hz) of the passband for whitened '
             'BLRMS, default: %(default)s',
    )
    parser.add_argument(
        '-S',
        '--sigma',
        type=float,
        default=2,
        help='significance threshold on whitened BLRMS, '
             'default: %(default)s',
    )
    parser.add_argument(
        '-c',
        '--main-channel',
        default='{IFO}:GDS-CALIB_STRAIN',
        help='name of main (h(t)) channel, default: %(default)s',
    )
    cli.add_frametype_option(
        parser,
        default='{IFO}_R',
    )
    parser.add_argument(
        '-o',
        '--output-dir',
        type=os.path.abspath,
        default=os.curdir,
        help='output directory for analysis, default: %(default)s',
    )
    parser.add_argument(
        '-w',
        '--omega-scans',
        type=int,
        metavar='NSCAN',
        default=0,
        help='generate a workflow of Omega scans for active '
             'scattering, this will launch automatically to '
             'condor, default: 0 (no Omega scans)',
    )
    parser.add_argument(
        '-v',
        '--verbose',
        action='store_true',
        default=False,
        help='log verbose output, default: %(default)s',
    )
    cli.add_nproc_option(
        parser,
    )

    # return the argument parser
    return parser


# -- main code block ----------------------------------------------------------

def main(args=None):
    """Run the primary scattering command-line tool
    """
    parser = create_parser()
    args = parser.parse_args(args=args)

    # set up logger
    logger = cli.logger(
        name=PROG.split('python -m ').pop(),
        level='DEBUG' if args.verbose else 'INFO',
    )

    # useful variables
    fthresh = (
        int(args.frequency_threshold) if args.frequency_threshold.is_integer()
        else args.frequency_threshold)
    multiplier = args.multiplier_for_threshold
    tstr = str(fthresh).replace('.', '_')
    gpsstr = '%s-%s' % (int(args.gpsstart), int(args.gpsend - args.gpsstart))
    args.optic = args.optic or list(OPTIC_MOTION_CHANNELS.keys())

    # go to working directory
    indir = os.getcwd()
    if not os.path.isdir(args.output_dir):
        os.makedirs(args.output_dir)
    os.chdir(args.output_dir)

    # set up output files
    summfile = '{}-SCATTERING_SUMMARY-{}.csv'.format(
        args.ifo, gpsstr)
    segfile = '{}-SCATTERING_SEGMENTS_{}_HZ-{}.h5'.format(
        args.ifo, tstr, gpsstr)

    # log start of process
    logger.info('{} Scattering {}-{}'.format(
        args.ifo, int(args.gpsstart), int(args.gpsend)))

    # -- get state segments -----------

    span = Segment(args.gpsstart, args.gpsend)

    # get segments
    if args.state_flag is not None:
        state = DataQualityFlag.query(
            args.state_flag, int(args.gpsstart), int(args.gpsend),
            url=DEFAULT_SEGMENT_SERVER,
        ).coalesce()
        statea = []
        padding = args.segment_start_pad + args.segment_end_pad
        for i, seg in enumerate(state.active):
            if abs(seg) > padding:
                statea.append(Segment(
                    seg[0] + args.segment_start_pad,
                    seg[1] - args.segment_end_pad,
                ))
            else:
                logger.debug(
                    "Segment length {} shorter than padding length {}, "
                    "skipping segment {}-{}".format(abs(seg), padding, *seg),
                )
        statea = SegmentList(statea)
        logger.debug("Downloaded %d segments for %s"
                     % (len(statea), args.state_flag))
    else:
        statea = SegmentList([span])
    livetime = float(abs(statea))
    logger.debug("Processing %.2f s of livetime" % livetime)

    # -- load h(t) --------------------

    args.main_channel = args.main_channel.format(IFO=args.ifo)
    logger.debug("Loading Omicron triggers for %s" % args.main_channel)

    if args.gpsstart >= 1230336018:  # Jan 1 2019
        ext = "h5"
        names = ["time", "frequency", "snr"]
        read_kw = {
            "columns": names,
            "selection": [
                "{0} < frequency < {1}".format(
                    args.fmin, multiplier * fthresh),
                ("time", in_segmentlist, statea),
            ],
            "format": "hdf5",
            "path": "triggers",
        }
    else:
        ext = "xml.gz"
        names = ['peak', 'peak_frequency', 'snr']
        read_kw = {
            "columns": names,
            "selection": [
                "{0} < peak_frequency < {1}".format(
                    args.fmin, multiplier * fthresh),
                ('peak', in_segmentlist, statea),
            ],
            "format": 'ligolw',
            "tablename": "sngl_burst",
        }

    fullcache = []
    for seg in statea:
        cache = gwtrigfind.find_trigger_files(
            args.main_channel, 'omicron', seg[0], seg[1], ext=ext,
        )
        if len(cache) == 0:
            warnings.warn(
                "No Omicron triggers found for %s in segment [%d .. %d)"
                % (args.main_channel, seg[0], seg[1]),
            )
            continue
        fullcache.extend(cache)

    # read triggers
    if fullcache:
        trigs = EventTable.read(fullcache, nproc=args.nproc, **read_kw)
    else:  # no files (no livetime?)
        trigs = EventTable(names=names)

    highsnrtrigs = trigs[trigs['snr'] >= 8]
    logger.debug("%d read" % len(trigs))

    # -- prepare HTML -----------------

    links = [
        '%d-%d' % (int(args.gpsstart), int(args.gpsend)),
        ('Parameters', '#parameters'),
        ('Segments', (
            ('State flag', '#state-flag'),
            ('Optical sensors', '#osems'),
            ('Transmons', '#transmons'),
        )),
    ]
    if args.omega_scans:
        links.append(('Scans', '#omega-scans'))
    (brand, class_) = htmlio.get_brand(args.ifo, 'Scattering', args.gpsstart)
    navbar = htmlio.navbar(links, class_=class_, brand=brand)
    page = htmlio.new_bootstrap_page(
        title='%s Scattering | %d-%d' % (
            args.ifo, int(args.gpsstart), int(args.gpsend)),
        navbar=navbar)
    page.div(class_='pb-2 mt-3 mb-2 border-bottom')
    page.h1('%s Scattering: %d-%d'
            % (args.ifo, int(args.gpsstart), int(args.gpsend)))
    page.div.close()  # pb-2 mt-3 mb-2 border-bottom
    page.h2('Parameters', class_='mt-4 mb-4', id_='parameters')
    page.div(class_='row')
    page.div(class_='col-md-9 col-sm-12')
    page.add(htmlio.parameter_table(
        start=int(args.gpsstart), end=int(args.gpsend), flag=args.state_flag))
    page.div.close()  # col-md-9 col-sm-12

    # link to summary files
    page.div(class_='col-md-3 col-sm-12')
    page.add(htmlio.download_btn(
        [('Segments (HDF)', segfile),
         ('Triggers (CSV)', summfile)],
        btnclass='btn btn-%s dropdown-toggle' % args.ifo.lower(),
    ))
    page.div.close()  # col-md-3 col-sm-12
    page.div.close()  # row

    # command-line
    page.h5('Command-line:')
    page.add(htmlio.get_command_line(about=False, prog=PROG))

    # section header
    page.h2('Segments', class_='mt-4', id_='segments')

    if statea:  # contextual information
        paper = markup.oneliner.a(
            'Accadia et al. (2010)', target='_blank', class_='alert-link',
            href='http://iopscience.iop.org/article/10.1088/0264-9381/27'
                 '/19/194011')
        msg = (
            "Segments marked \"optical sensors\" below show evidence of beam "
            "scattering between {0} and {1} Hz based on the velocity of optic "
            "motion, with fringe frequencies projected using equation (3) of "
            "{2}. Segments marked \"transmons\" are based on whitened, "
            "band-limited RMS trends of transmon sensors. In both cases, "
            "yellow panels denote weak evidence for scattering, while red "
            "panels denote strong evidence."
         ).format(args.fmin, multiplier * fthresh, str(paper))
        page.add(htmlio.alert(msg, context=args.ifo.lower()))
    else:  # null segments
        page.add(htmlio.alert('No active analysis segments were found',
                              context='warning', dismiss=False))

    # record state segments
    if args.state_flag is not None:
        page.h3('State flag', class_='mt-3', id_='state-flag')
        page.div(id_='accordion1')
        page.add(htmlio.write_flag_html(
            state, span, 'state', parent='accordion1', context='success',
            plotdir='', facecolor=(0.2, 0.8, 0.2), edgecolor='darkgreen',
            known={'facecolor': 'red', 'edgecolor': 'darkred', 'height': 0.4}))
        page.div.close()

    # -- find scattering evidence -----

    # read data for OSEMs and transmons
    osems = ['%s:%s' % (args.ifo, c) for optic in args.optic for
             c in OPTIC_MOTION_CHANNELS[optic]]
    transmons = ['%s:%s' % (args.ifo, c) for c in TRANSMON_CHANNELS]
    allchannels = osems + transmons

    logger.info("Reading all timeseries data")
    alldata = []
    n = len(statea)
    for i, seg in enumerate(statea):
        msg = "{0}/{1} {2}:".rjust(30).format(
            str(i + 1).rjust(len(str(n))),
            n,
            str(seg),
        ) if args.verbose else False
        alldata.append(
            get_data(allchannels, seg[0], seg[1],
                     frametype=args.frametype.format(IFO=args.ifo),
                     verbose=msg, nproc=args.nproc).resample(128))
    try:  # ensure that only available channels are analyzed
        osems = list(
            set(alldata[0].keys()) & set(alldata[-1].keys()) & set(osems))
        transmons = list(
            set(alldata[0].keys()) & set(alldata[-1].keys()) & set(transmons))
    except IndexError:
        osems = []
        transmons = []

    # initialize scattering segments
    scatter_segments = DataQualityDict()
    actives = SegmentList()

    # scattering based on OSEM velocity
    if statea:
        page.h3('Optical sensors (OSEMs)', class_='mt-3', id_='osems')
        page.div(id_='osems-group')
    logger.info('Searching for scatter based on OSEM velocity')

    for i, channel in enumerate(sorted(osems)):
        logger.info("-- Processing %s --" % channel)
        chanstr = re.sub('[:-]', '_', channel).replace('_', '-', 1)
        optic = channel.split('-')[1].split('_')[0]
        flag = '%s:DCH-%s_SCATTERING_GE_%s_HZ:1' % (args.ifo, optic, tstr)
        scatter_segments[channel] = DataQualityFlag(
            flag,
            isgood=False,
            description="Evidence for scattering above {0} Hz from {1} in "
                        "{2}".format(fthresh, optic, channel),
        )
        # set up plot(s)
        plot = Plot(figsize=[12, 12])
        axes = {}
        axes['position'] = plot.add_subplot(
            411, xscale='auto-gps', xlabel='')
        axes['fringef'] = plot.add_subplot(
            412, sharex=axes['position'], xlabel='')
        axes['triggers'] = plot.add_subplot(
            413, sharex=axes['position'], xlabel='')
        axes['segments'] = plot.add_subplot(
            414, projection='segments', sharex=axes['position'])
        plot.subplots_adjust(bottom=.07, top=.95)
        fringecolors = [None] * len(FREQUENCY_MULTIPLIERS)
        histdata = dict((x, numpy.ndarray((0,))) for
                        x in FREQUENCY_MULTIPLIERS)
        linecolor = None
        # loop over state segments and find scattering fringes
        for j, seg in enumerate(statea):
            logger.debug("Processing segment [%d .. %d)" % seg)
            ts = alldata[j][channel]
            # get raw data and plot
            line = axes['position'].plot(ts, color=linecolor)[0]
            linecolor = line.get_color()
            # get fringe frequency and plot
            fringef = get_fringe_frequency(ts, multiplier=1)
            for k, m in list(enumerate(FREQUENCY_MULTIPLIERS))[::-1]:
                fm = fringef * m
                line = axes['fringef'].plot(
                    fm, color=fringecolors[k],
                    label=(j == 0 and r'$f\times%d$' % m or None))[0]
                fringecolors[k] = line.get_color()
                histdata[m] = numpy.resize(
                    histdata[m], (histdata[m].size + fm.size,))
                histdata[m][-fm.size:] = fm.value
            # get segments and plot
            scatter = get_segments(
                fringef * multiplier,
                fthresh,
                name=flag,
                pad=args.segment_padding
            )
            axes['segments'].plot(
                scatter, facecolor='red', edgecolor='darkred',
                known={'alpha': 0.6, 'facecolor': 'lightgray',
                       'edgecolor': 'gray', 'height': 0.4},
                height=0.8, y=0, label=' ',
            )
            scatter_segments[channel] += scatter
            logger.debug(
                "    Found %d scattering segments" % (len(scatter.active)))
        logger.debug("Completed channel %s, found %d segments in total"
                     % (channel, len(scatter_segments[channel].active)))

        # calculate efficiency and deadtime of veto
        deadtime = abs(scatter_segments[channel].active)
        try:
            deadtimepc = deadtime / livetime * 100
        except ZeroDivisionError:
            deadtimepc = 0.
        logger.info("Deadtime: %.2f%% (%.2f/%ds)"
                    % (deadtimepc, deadtime, livetime))
        efficiency = in_segmentlist(highsnrtrigs[names[0]],
                                    scatter_segments[channel].active).sum()
        try:
            efficiencypc = efficiency / len(highsnrtrigs) * 100
        except ZeroDivisionError:
            efficiencypc = 0.
        logger.info("Efficiency (SNR>=8): %.2f%% (%d/%d)"
                    % (efficiencypc, efficiency, len(highsnrtrigs)))
        if deadtimepc == 0.:
            effdt = 0
        else:
            effdt = efficiencypc/deadtimepc
        logger.info("Efficiency/Deadtime: %.2f" % effdt)

        if abs(scatter_segments[channel].active):
            actives.extend(scatter_segments[channel].active)

        # finalize plot
        logger.debug("Plotting")
        name = texify(channel)
        axes['position'].set_title("Scattering evidence in %s" % name)
        axes['position'].set_xlabel('')
        axes['position'].set_ylabel(r'Position [$\mu$m]')
        axes['position'].text(
            0.01, 0.95, 'Optic position',
            transform=axes['position'].transAxes, va='top', ha='left',
            bbox={'edgecolor': 'none', 'facecolor': 'white', 'alpha': .5})
        axes['fringef'].plot(
            span, [fthresh, fthresh], 'k--')
        axes['fringef'].set_xlabel('')
        axes['fringef'].set_ylabel(r'Frequency [Hz]')
        axes['fringef'].yaxis.tick_right()
        axes['fringef'].yaxis.set_label_position("right")
        axes['fringef'].set_ylim(0, multiplier * fthresh)
        axes['fringef'].text(
            0.01, 0.95, 'Calculated fringe frequency',
            transform=axes['fringef'].transAxes, va='top', ha='left',
            bbox={'edgecolor': 'none', 'facecolor': 'white', 'alpha': .5})
        handles, labels = axes['fringef'].get_legend_handles_labels()
        axes['fringef'].legend(handles[::-1], labels[::-1], loc='upper right',
                               borderaxespad=0, bbox_to_anchor=(-0.01, 1.),
                               handlelength=1)

        axes['triggers'].scatter(
            trigs[names[0]],
            trigs[names[1]],
            c=trigs[names[2]],
            edgecolor='none',
        )
        name = texify(args.main_channel)
        axes['triggers'].text(
            0.01, 0.95,
            '%s event triggers (Omicron)' % name,
            transform=axes['triggers'].transAxes, va='top', ha='left',
            bbox={'edgecolor': 'none', 'facecolor': 'white', 'alpha': .5})
        axes['triggers'].set_ylabel('Frequency [Hz]')
        axes['triggers'].set_ylim(args.fmin, multiplier * fthresh)
        axes['triggers'].colorbar(cmap='YlGnBu', clim=(3, 100), norm='log',
                                  label='Signal-to-noise ratio')
        axes['segments'].set_ylim(-.55, .55)
        axes['segments'].text(
            0.01, 0.95,
            r'Time segments with $f\times%d > %.2f$ Hz' % (
                multiplier, fthresh),
            transform=axes['segments'].transAxes, va='top', ha='left',
            bbox={'edgecolor': 'none', 'facecolor': 'white', 'alpha': .5})
        for ax in axes.values():
            ax.set_epoch(int(args.gpsstart))
            ax.set_xlim(*span)
        png = '%s_SCATTERING_%s_HZ-%s.png' % (chanstr, tstr, gpsstr)
        try:
            plot.save(png)
        except OverflowError as e:
            warnings.warn(str(e))
            plot.axes[1].set_ylim(0, multiplier * fthresh)
            plot.refresh()
            plot.save(png)
        plot.close()
        logger.debug("%s written." % png)

        # make histogram
        histogram = Plot(figsize=[12, 6])
        ax = histogram.gca()
        hrange = (0, multiplier * fthresh)
        for m, color in list(zip(histdata, fringecolors))[::-1]:
            if histdata[m].size:
                ax.hist(
                    histdata[m], facecolor=color, alpha=.6, range=hrange,
                    bins=50, histtype='stepfilled', label=r'$f\times%d$' % m,
                    cumulative=-1, weights=ts.dx.value, bottom=1e-100,
                    log=True)
            else:
                ax.plot(histdata[m], color=color, label=r'$f\times%d$' % m)
                ax.set_yscale('log')
        ax.set_ylim(.01, float(livetime))
        ax.set_ylabel('Time with fringe above frequency [s]')
        ax.set_xlim(*hrange)
        ax.set_xlabel('Frequency [Hz]')
        ax.set_title(axes['position'].get_title())
        handles, labels = ax.get_legend_handles_labels()
        ax.legend(handles[::-1], labels[::-1], loc='upper right')
        hpng = '%s_SCATTERING_HISTOGRAM-%s.png' % (chanstr, gpsstr)
        histogram.save(hpng)
        histogram.close()
        logger.debug("%s written." % hpng)

        # write HTML
        if deadtime != 0 and effdt > 2:
            context = 'danger'
        elif ((deadtime != 0 and effdt < 2) or
              (histdata[multiplier].size and
               histdata[multiplier].max() >=
                  fthresh/2.)):
            context = 'warning'
        else:
            continue
        page.div(class_='card border-%s mb-1 shadow-sm' % context)
        page.div(class_='card-header text-white bg-%s' % context)
        page.a(channel, class_='collapsed card-link cis-link',
               href='#osem%s' % i, **{'data-toggle': 'collapse'})
        page.div.close()  # card-header
        page.div(id_='osem%s' % i, class_='collapse',
                 **{'data-parent': '#osems-group'})
        page.div(class_='card-body')
        page.div(class_='row')
        img = htmlio.FancyPlot(
            png, caption=SCATTER_CAPTION.format(CHANNEL=channel))
        page.div(class_='col-md-10 offset-md-1')
        page.add(htmlio.fancybox_img(img))
        page.div.close()  # col-md-10 offset-md-1
        himg = htmlio.FancyPlot(
            hpng, caption=HIST_CAPTION.format(CHANNEL=channel))
        page.div(class_='col-md-10 offset-md-1')
        page.add(htmlio.fancybox_img(himg))
        page.div.close()  # col-md-10 offset-md-1
        page.div.close()  # row
        segs = StringIO()
        if deadtime:
            page.p("%d segments were found predicting a scattering fringe "
                   "above %.2f Hz." % (
                       len(scatter_segments[channel].active),
                       fthresh))
            page.table(class_='table table-sm table-hover')
            page.tbody()
            page.tr()
            page.th('Deadtime')
            page.td('%.2f/%d seconds' % (deadtime, livetime))
            page.td('%.2f%%' % deadtimepc)
            page.tr.close()
            page.tr()
            page.th('Efficiency<br><small>(SNR&ge;8 and '
                    '%.2f Hz</sub>&ltf<sub>peak</sub>&lt;%.2f Hz)</small>'
                    % (args.fmin, multiplier * fthresh))
            page.td('%d/%d events' % (efficiency, len(highsnrtrigs)))
            page.td('%.2f%%' % efficiencypc)
            page.tr.close()
            page.tr()
            page.th('Efficiency/Deadtime')
            page.td()
            page.td('%.2f' % effdt)
            page.tr.close()
            page.tbody.close()
            page.table.close()
            scatter_segments[channel].active.write(segs, format='segwizard',
                                                   coltype=float)
            page.pre(segs.getvalue())
        else:
            page.p("No segments were found with scattering above %.2f Hz."
                   % fthresh)
        page.div.close()  # card-body
        page.div.close()  # collapse
        page.div.close()  # card

    if statea:  # close accordion
        page.div.close()  # osems-group

    # scattering based on transmon BLRMS
    if statea:
        page.h3('Transmons', class_='mt-3', id_='transmons')
        page.div(id_='transmons-group')
    logger.info('Searching for scatter based on band-limited RMS of transmons')

    for i, channel in enumerate(sorted(transmons)):
        logger.info("-- Processing %s --" % channel)
        optic = channel.split('-')[1][:6]
        flag = '%s:DCH-%s_SCATTERING_BLRMS:1' % (args.ifo, optic)
        scatter_segments[channel] = DataQualityFlag(
            flag,
            isgood=False,
            description="Evidence for scattering from whitened, band-limited "
                        "RMS trends of {0}".format(channel),
        )

        # loop over state segments and compute BLRMS
        for j, seg in enumerate(statea):
            logger.debug("Processing segment [%d .. %d)" % seg)
            wblrms = get_blrms(
                alldata[j][channel],
                flow=args.bandpass_flow,
                fhigh=args.bandpass_fhigh,
            )
            scatter = get_segments(
                wblrms,
                numpy.mean(wblrms) + args.sigma * numpy.std(wblrms),
                name=flag,
            )
            scatter_segments[channel] += scatter
            logger.debug(
                "    Found %d scattering segments" % (len(scatter.active)))
        logger.debug("Completed channel %s, found %d segments in total"
                     % (channel, len(scatter_segments[channel].active)))

        # calculate efficiency and deadtime of veto
        deadtime = abs(scatter_segments[channel].active)
        try:
            deadtimepc = deadtime / livetime * 100
        except ZeroDivisionError:
            deadtimepc = 0.
        logger.info("Deadtime: %.2f%% (%.2f/%ds)"
                    % (deadtimepc, deadtime, livetime))
        highsnrtrigs = trigs[trigs['snr'] <= 200]
        efficiency = in_segmentlist(highsnrtrigs[names[0]],
                                    scatter_segments[channel].active).sum()
        try:
            efficiencypc = efficiency / len(highsnrtrigs) * 100
        except ZeroDivisionError:
            efficiencypc = 0.
        logger.info("Efficiency (SNR>=8): %.2f%% (%d/%d)"
                    % (efficiencypc, efficiency, len(highsnrtrigs)))
        if deadtimepc == 0.:
            effdt = 0
        else:
            effdt = efficiencypc/deadtimepc
        logger.info("Efficiency/Deadtime: %.2f" % effdt)

        if abs(scatter_segments[channel].active):
            actives.extend(scatter_segments[channel].active)

        # write HTML
        if deadtime != 0 and effdt > 2:
            context = 'danger'
        elif deadtime != 0 and effdt < 2:
            context = 'warning'
        else:
            continue
        page.add(htmlio.write_flag_html(
            scatter_segments[channel], span, i, parent='transmons-group',
            title=channel, context=context, plotdir=''))

    if statea:  # close accordion
        page.div.close()  # transmons-group

    actives = actives.coalesce()  # merge contiguous segments
    if statea and not actives:
        page.add(htmlio.alert(
            'No evidence of scattering found in the channels analyzed',
            context=args.ifo.lower(), dismiss=False))

    # identify triggers during active segments
    logger.debug('Writing a summary CSV record')
    ind = [i for i, trigtime in enumerate(highsnrtrigs[names[0]])
           if trigtime in actives]
    gps = highsnrtrigs[names[0]][ind]
    freq = highsnrtrigs[names[1]][ind]
    snr = highsnrtrigs[names[2]][ind]
    segs = [y for x in gps for y in actives if x in y]
    table = EventTable(
        [gps, freq, snr, [seg[0] for seg in segs], [seg[1] for seg in segs]],
        names=('trigger_time', 'trigger_frequency', 'trigger_snr',
               'segment_start', 'segment_end'))
    logger.info('The following {} triggers fell within active scattering '
                'segments:\n\n'.format(len(table)))
    print(table)
    print('\n\n')
    table.write(summfile, overwrite=True)

    # -- launch omega scans -----------

    nscans = min(args.omega_scans, len(table))
    if nscans > 0:
        # launch scans
        scandir = 'scans'
        ind = random.sample(range(0, len(table)), nscans)
        omegatimes = [str(t) for t in table['trigger_time'][ind]]
        logger.debug('Collected {} event times to omega scan: {}'.format(
            nscans, ', '.join(omegatimes)))
        logger.info('Creating workflow for omega scans')
        flags = batch.get_command_line_flags(
            ifo=args.ifo, ignore_state_flags=True)
        condorcmds = batch.get_condor_arguments(timeout=4, gps=args.gpsstart)
        batch.generate_dag(omegatimes, flags=flags, submit=True,
                           outdir=scandir, condor_commands=condorcmds)
        logger.info('Launched {} omega scans to condor'.format(nscans))
        # render HTML
        page.h2('Omega scans', class_='mt-4', id_='omega-scans')
        msg = (
            'The following event times correspond to significant Omicron '
            'triggers that occur during the scattering segments found above. '
            'To compare these against fringe frequency projections, please '
            'use the "simple scattering" module:',
            markup.oneliner.pre(
                '$ python -m gwdetchar.scattering.simple --help',
            ),
        )
        page.add(htmlio.alert(msg, context=args.ifo.lower()))
        page.add(htmlio.scaffold_omega_scans(
            omegatimes, args.main_channel, scandir=scandir))
    elif args.omega_scans:
        logger.info('No events found during active scattering segments')

    # -- finalize ---------------------

    # write segments
    scatter_segments.write(segfile, path="segments", overwrite=True)
    logger.debug("%s written" % segfile)

    # write HTML
    htmlio.close_page(page, 'index.html')
    logger.info("-- index.html written, all done --")

    # return to original directory
    os.chdir(indir)


# -- run from command-line ----------------------------------------------------

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Copyright (C) Alex Urban (2020)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gwdetchar.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for the `gwdetchar.scattering` command-line interface
"""

import os
import glob
import numpy
import shutil

from unittest import mock

from gwpy.segments import (
    Segment,
    SegmentList,
    DataQualityFlag,
)
from gwpy.timeseries import (
    TimeSeries,
    TimeSeriesDict,
)

from .. import __main__ as scattering_cli

__author__ = 'Alex Urban <alexander.urban@ligo.org>'

IFO = 'H1'

# -- test data

# The unit test below captures use cases by simulating
#     (1) empty livetime from requested segments, and
#     (2) a scattering fringe in h(t) predicted by only one optic.

DURATION = 608
FREQ = 1 / 10
SAMPLE = 4096

EMPTY_FLAG = DataQualityFlag(
    name="{}:DCH-TEST_FLAG:1".format(IFO),
    active=SegmentList([
        Segment(0, 25),
    ]),
    known=SegmentList([
        Segment(0, DURATION),
    ]),
)

TIMES = numpy.arange(0, DURATION, 1 / SAMPLE)
PHASE = 42 * numpy.sin(2 * numpy.pi * FREQ * TIMES) / (2 * numpy.pi * FREQ)
SCATTER = TimeSeries(
    (numpy.sin(numpy.pi * TIMES / DURATION)
     * numpy.cos(2 * numpy.pi * PHASE)),
    sample_rate=SAMPLE,
).highpass(10)

OSEMS = [chan for group in scattering_cli.OPTIC_MOTION_CHANNELS.values()
         for chan in group]
AUX = TimeSeriesDict({
    **{
        ':'.join([IFO, chan]): TimeSeries(
            numpy.random.normal(loc=1, scale=1e-3, size=SCATTER.size),
            sample_rate=SAMPLE,
            name=':'.join([IFO, chan]),
        ).crop(4, DURATION - 4) for chan in OSEMS[1::]
    },
    **{
        ':'.join([IFO, chan]): TimeSeries(
            numpy.random.normal(loc=1, scale=1.5, size=SCATTER.size),
            sample_rate=SAMPLE,
            name=':'.join([IFO, chan]),
        ).inject(SCATTER).crop(
            4, DURATION - 4) for chan in scattering_cli.TRANSMON_CHANNELS},
})

PHASE = PHASE[4 * SAMPLE:-4 * SAMPLE]
AUX['{}:SUS-ITMX_R0_DAMP_L_IN1_DQ'.format(IFO)] += 1.064 * PHASE / 2


# -- cli tests ----------------------------------------------------------------

@mock.patch(
    'gwpy.segments.DataQualityFlag.query',
    return_value=EMPTY_FLAG,
)
def test_main_no_livetime(flag, caplog, tmpdir):
    outdir = str(tmpdir)
    args = [  # command-line arguments
        '-i', IFO,
        '4', str(DURATION - 4),
        '--state-flag', '{}:DCH-TEST_FLAG:1'.format(IFO),
        '--multiplier-for-threshold', '1',
        '--output-dir', outdir,
        '--omega-scans', '5',
        '--verbose',
    ]
    # test command-line tool with no livetime
    scattering_cli.main(args)
    assert ("Segment length 25 shorter than padding length 60.0, skipping "
            "segment 0-25" in caplog.text)
    assert "Downloaded 0 segments for H1:DCH-TEST_FLAG:1" in caplog.text
    assert "No events found during active scattering segments" in caplog.text
    assert not os.path.exists(os.path.join(outdir, 'scans'))
    with open(os.path.join(outdir, "index.html"), 'r') as f:
        assert "No active analysis segments were found" in f.read()
    # clean up
    shutil.rmtree(outdir, ignore_errors=True)


@mock.patch(
    'gwtrigfind.find_trigger_files',
    return_value=[],
)
@mock.patch(
    'gwdetchar.scattering.__main__.get_data',
    return_value=AUX,
)
def test_main(cache, data, caplog, tmpdir, recwarn):
    outdir = str(tmpdir)
    args = [  # command-line arguments
        '-i', IFO,
        '4', str(DURATION - 4),
        '--multiplier-for-threshold', '1',
        '--output-dir', outdir,
        '--verbose',
    ]
    # test command-line tool
    scattering_cli.main(args)
    hdf = "{0}-SCATTERING_SEGMENTS_15_HZ-4-{1}.h5".format(IFO, DURATION - 8)
    assert "Processing %.2f s of livetime" % (DURATION - 8) in caplog.text
    assert "Searching for scatter based on OSEM velocity" in caplog.text
    assert ("Searching for scatter based on band-limited RMS of transmons"
            in caplog.text)
    assert "Writing a summary CSV record" in caplog.text
    assert ("The following 0 triggers fell within active scattering segments:"
            in caplog.text)
    assert "{0} written".format(hdf) in caplog.text
    assert "-- index.html written, all done --" in caplog.text
    for channel in AUX.keys():
        assert "-- Processing {0} --".format(channel) in caplog.text
        assert "Completed channel {0}".format(channel) in caplog.text
    # test output data products
    assert glob.glob(os.path.join(outdir, '*.png'))
    assert len(glob.glob(os.path.join(outdir, '*.csv'))) == 1
    assert len(glob.glob(os.path.join(outdir, '*.html'))) == 1
    assert os.path.exists(os.path.join(outdir, hdf))
    assert not os.path.exists(os.path.join(outdir, 'scans'))
    # reject warnings due to no Omicron triggers
    recwarn.pop(RuntimeWarning)
    recwarn.pop(UserWarning)
    # clean up
    shutil.rmtree(outdir, ignore_errors=True)



# -*- coding: utf-8 -*-
# Copyright (C) Alex Urban (2019)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gwdetchar.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for `gwdetchar.lasso.plot`
"""

import numpy
import os
import pytest
import shutil

from gwpy.timeseries import TimeSeries

from matplotlib import (
    use,
    rcParams,
    rcParamsDefault,
)
use('Agg')

# backend-dependent import
from .. import plot  # noqa: E402

__author__ = 'Alex Urban <alexander.urban@ligo.org>'


# -- make sure plots run end-to-end -------------------------------------------

def test_configure_mpl_tex():
    plot.configure_mpl_tex()
    assert os.environ['HOME'] == os.environ['MPLCONFIGDIR']
    assert rcParams['ps.useafm'] is True
    assert rcParams['pdf.use14corefonts'] is True
    assert rcParams['text.usetex'] is True
    rcParams.update(rcParamsDefault)


def test_save_figure(tmpdir):
    base = str(tmpdir)
    series = TimeSeries(numpy.random.normal(loc=0, scale=1, size=24*60),
                        sample_rate=60, unit='Mpc', name='X1:TEST')
    fig = series.plot()
    tsplot = plot.save_figure(fig, os.path.join(base, 'test.png'))
    assert tsplot == os.path.join(base, 'test.png')

    # no-directory should raise warning
    with pytest.warns(UserWarning) as record:
        noneplot = plot.save_figure(fig, os.path.join('tgpflk', 'test.png'))
    assert noneplot is None
    assert len(record.list) == 1
    assert 'Error saving' in str(record.list[0].message)

    # remove base directory
    shutil.rmtree(base, ignore_errors=True)



# coding=utf-8
# Copyright (C) LIGO Scientific Collaboration (2019)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Check whether state records have changed between two reference times
"""

import gwdatafind
import numpy
import os
import re
import sys

from gwpy.io import gwf as io_gwf
from gwpy.table import EventTable

from . import cli
from .io.datafind import get_data

__author__ = 'Alex Urban <alexander.urban@ligo.org>'
__credits__ = ('Andrew Lundgren <andrew.lundgren@ligo.org>, '
               'Joshua Smith <joshua.smith@ligo.org>, '
               'Duncan Macleod <duncan.macleod@ligo.org>')

# set up logger
PROG = ('python -m gwdetchar.conlog' if sys.argv[0].endswith('.py')
        else os.path.basename(sys.argv[0]))
LOGGER = cli.logger(name=PROG.split('python -m ').pop())


# -- utilities ----------------------------------------------------------------

def _discover_data_source(obs, frametype, start, end, preview):
    """Determine filepaths to local gravitational-wave frame files
    """
    # get paths to frame files
    cache1 = gwdatafind.find_urls(
        obs,
        frametype,
        start - preview,
        start,
    )
    cache2 = gwdatafind.find_urls(
        obs,
        frametype,
        end,
        end + 1,
    )
    return (cache1, cache2)


def _discover_data(channels, caches, start, end, preview, nproc=1):
    """Load timeseries data from local gravitational-wave frame files
    """
    (cache1, cache2) = caches
    # get data from frames
    data1 = get_data(
        channels,
        start=start - preview,
        end=start,
        source=cache1,
        nproc=nproc,
        verbose='Reading initial data:'.rjust(30),
    )
    data2 = get_data(
        channels,
        start=end,
        end=end + 1,
        source=cache2,
        nproc=nproc,
        verbose='Reading final data:'.rjust(30),
    )
    return (data1, data2)


def _get_available_channels(caches):
    """Determine available channels from local gravitational-wave frame files
    """
    (cache1, cache2) = caches
    try:
        available = (
            set(io_gwf.iter_channel_names(cache1[-1])) &
            set(io_gwf.iter_channel_names(cache2[0]))
        )
    except (IndexError, TypeError):
        raise RuntimeError("Could not find data in the time range requested")
    return available


# -- parse command-line -------------------------------------------------------

def create_parser():
    """Create a command-line parser for this entry point
    """
    # initialize argument parser
    parser = cli.create_parser(
        prog=PROG,
        description=__doc__,
    )

    # set arguments from the cli module
    cli.add_gps_start_stop_arguments(parser)
    cli.add_ifo_option(parser)
    cli.add_frametype_option(parser,
                             help='the frametype name, defaults to second '
                                  'trends for the selected interferometer')
    cli.add_nproc_option(parser)

    # other optional arguments
    parser.add_argument(
        '-o',
        '--output',
        default='changes.csv',
        help='Path to output data file, default: %(default)s',
    )
    parser.add_argument(
        '-c',
        '--channels',
        default=None,
        required=False,
        help='file containing columnar list of channels to '
             'process, default is to find all relevant channels '
             'from frames',
    )
    parser.add_argument(
        '-s',
        '--search',
        nargs='*',
        default=[],
        help='process channels matching these regex patterns, '
             'can be given multiple times, default is to analyze '
             'all relevant channels from frames',
    )
    parser.add_argument(
        '-p',
        '--preview',
        default=10,
        type=int,
        help='Time (seconds) over which to test that channel is '
             'normally kept constant, default: %(default)s',
    )

    # return the argument parser
    return parser


# -- main code block ----------------------------------------------------------

def main(args=None):
    """Run the Conlog command-line tool
    """
    parser = create_parser()
    args = parser.parse_args(args=args)

    # get IFO and frametype
    ifo = args.ifo.upper()
    obs = ifo[0]
    frametype = args.frametype or '{}_T'.format(ifo)
    preview_time = max(1, args.preview)

    # get paths to frame files
    (cache1, cache2) = _discover_data_source(
        obs,
        frametype,
        args.gpsstart,
        args.gpsend,
        preview_time,
    )

    # get list of channels to analyze
    LOGGER.info('Determining channels to analyze')
    available = _get_available_channels((cache1, cache2))
    if args.channels:
        reqchannels = set(numpy.loadtxt(args.channels, dtype=str, ndmin=1))
        channels = list(reqchannels & available)
    else:
        channels = [x for x in available if x.endswith('.mean')]
    if args.search:  # if requested, get channels matching regex patterns
        re_requested = re.compile('({})'.format('|'.join(args.search)))
        channels = [x for x in channels if re_requested.search(x)]
    LOGGER.debug('Found {} channels in frames'.format(len(channels)))

    # get data from frames
    (data1, data2) = _discover_data(
        channels,
        (cache1, cache2),
        args.gpsstart,
        args.gpsend,
        preview_time,
        nproc=args.nproc,
    )

    # initialize columns
    LOGGER.debug('Analyzing {} channels'.format(len(data1)))
    changes = []
    value1 = []
    value2 = []
    diff = []

    # identify channels
    for channel in data1:
        xoft1 = data1[channel].value
        xoft2 = data2[channel].value
        if numpy.any(numpy.diff(xoft1) != 0):
            continue
        if xoft1[-1] == xoft2[0]:
            continue
        changes.append(channel)
        value1.append(xoft1[-1])
        value2.append(xoft2[0])
        diff.append(xoft2[0] - xoft1[-1])

    # record output
    LOGGER.debug('Analysis complete')
    table = EventTable([changes, value1, value2, diff],
                       names=('channel', 'initial_value',
                              'final_value', 'difference'))

    # log output
    LOGGER.info('The following {0} channels record a state '
                'change between {1} and {2}:\n\n'.format(
                    len(changes),
                    args.gpsstart,
                    args.gpsend,
                ))
    print(table)
    print('\n\n')

    # save output
    table.write(args.output, overwrite=True)
    LOGGER.info('Output written to {}'.format(args.output))


# -- run code -----------------------------------------------------------------

if __name__ == "__main__":
    main()



# coding=utf-8
# Copyright (C) Duncan Macleod (2015)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Utility methods
"""

import re
import sys
from io import StringIO
from functools import partial
from html.parser import HTMLParser

import numpy

from gwpy.table import EventTable

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__credits__ = 'Alex Urban <alexander.urban@ligo.org>'


# -- class for HTML parsing ---------------------------------------------------

class GWHTMLParser(HTMLParser):
    """See https://docs.python.org/3/library/html.parser.html.
    """
    def handle_starttag(self, tag, attrs):
        print("Start tag:", tag)
        attrs.sort()
        for attr in attrs:
            print("attr:", attr)

    def handle_endtag(self, tag):
        print("End tag:", tag)

    def handle_data(self, data):
        print("Data:", data)

    def handle_decl(self, data):
        print("Decl:", data)


parser = GWHTMLParser()


# -- utilities ----------------------------------------------------------------

def parse_html(html):
    """Parse a string containing raw HTML code
    """
    stdout = sys.stdout
    sys.stdout = StringIO()
    if sys.version_info.major < 3:
        parser.feed(html.decode('utf-8', 'ignore'))
    else:
        parser.feed(html)
    output = sys.stdout.getvalue()
    sys.stdout = stdout
    return output


def natural_sort(ls, key=str):
    """Sort a list the way that humans expect.

    This differs from the built-in `sorted` method by building a custom
    key to sort numeric parts by value, not by character content, so that
    '2' gets sorted before '10', for example.

    Parameters
    ----------
    ls : `iterable`
        iterable to sort
    key : `callable`
        sorting key

    Returns
    -------
    sorted : `list`
        a sorted version of the input list
    """
    ls = list(ls)
    k = list(map(key, ls)) if key else ls

    def convert(text):
        return int(text) if text.isdigit() else text

    def alphanum_key(key):
        return [convert(c) for c in
                re.split('([0-9]+)', k[ls.index(key)])]

    return sorted(ls, key=alphanum_key)


def table_from_segments(flagdict, sngl_burst=False, snr=10., frequency=100.):
    """Build an `EventTable` from a `DataQualityDict`
    """
    rows = []
    if sngl_burst:
        names = ("peak", "peak_frequency", "snr", "channel")

        def row(seg, channel):
            a, b = map(float, seg)
            return a, frequency, snr, channel
    else:
        names = ("time", "frequency", "start_time", "end_time",
                 "snr", "channel")

        def row(seg, channel):
            a, b = map(float, seg)
            return a, frequency, a, b, snr, channel

    for name, flag in flagdict.items():
        rows.extend(map(partial(row, channel=name), flag.active))
    table = EventTable(rows=rows or None, names=names)
    if sngl_burst:  # add tablename for GWpy's ligolw writer
        table.meta["tablename"] = "sngl_burst"
    return table


def table_from_times(times, names=("time", "frequency", "snr"),
                     snr=10., frequency=100., **kwargs):
    """Build an `EventTable` from a `DataQualityDict`

    Parameters
    ----------
    times : `numpy.ndarray`
        a 1D array of GPS times

    names : `tuple`, `list`, optional
        the list of names to use

    snr : `float`, optional
        the SNR to assign to all 'event triggers'

    frequency : `float`, optional
        the frequency to assign to all 'event triggers'

    **kwargs
        other keyword arguments to pass to the `EventTable` constructor

    Returns
    -------
    table : `~gwpy.table.EventTable`
        a new table filled with events at the given times
    """
    farr = numpy.ones_like(times) * frequency
    sarr = numpy.ones_like(times) * snr
    return EventTable([times, farr, sarr], names=names, **kwargs)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of the GW DetChar python (gwdetchar) package.
#
# gwdetchar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# gwdetchar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gwdetchar.  If not, see <http://www.gnu.org/licenses/>.

"""Setup the gwdetchar package
"""

from setuptools import setup

import versioneer

# versioneer
version = versioneer.get_version()
cmdclass = versioneer.get_cmdclass()

# import sphinx commands
try:
    from sphinx.setup_command import BuildDoc
except ImportError:
    pass
else:
    cmdclass['build_sphinx'] = BuildDoc

# run setup
# NOTE: all other metadata and options come from setup.cfg
setup(
    version=version,
    cmdclass=cmdclass,
)



# -*- coding: utf-8 -*-
# Copyright (C) Joshua Smith (2016-)
#
# This file is part of the GW DetChar python package.
#
# gwdetchar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# gwdetchar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gwdetchar.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for `gwdetchar.io.ligolw`
"""

import pytest

import numpy
from numpy.testing import (assert_array_equal, assert_allclose)

from gwpy.segments import (Segment, SegmentList)
from gwpy.testing.utils import assert_segmentlist_equal

lsctables = pytest.importorskip("glue.ligolw.lsctables")
ligolw = pytest.importorskip("gwdetchar.io.ligolw")


def test_new_table():
    tab = ligolw.new_table('sngl_burst')
    assert isinstance(tab, lsctables.SnglBurstTable)
    assert sorted(tab.columnnames) == sorted(
        lsctables.SnglBurstTable.validcolumns.keys())

    cols = ['peak_time', 'peak_time_ns', 'snr']
    tab = ligolw.new_table(lsctables.SnglBurstTable, columns=cols)
    assert tab.columnnames == cols


def test_sngl_burst_from_times():
    numpy.random.seed(0)
    times = numpy.random.random(4) * 100.
    tab = ligolw.sngl_burst_from_times(times)
    assert isinstance(tab, lsctables.SnglBurstTable)
    assert len(tab) == 4
    nanosec, sec = numpy.modf(times)
    assert_array_equal(tab.getColumnByName('peak_time').asarray(), sec)
    assert_allclose(tab.getColumnByName('peak_time_ns').asarray(),
                    nanosec * 1e9)


def test_segments_from_sngl_burst():
    tab = ligolw.sngl_burst_from_times([1, 4, 7, 10], channel='test')
    segs = ligolw.segments_from_sngl_burst(tab, 1)
    assert_segmentlist_equal(segs['test'].active, SegmentList([
        Segment(0, 2), Segment(3, 5), Segment(6, 8), Segment(9, 11),
    ]))


def test_table_to_document():
    from glue.ligolw.ligolw import Document
    tab = ligolw.new_table('sngl_burst')
    xmldoc = ligolw.table_to_document(tab)
    assert isinstance(xmldoc, Document)
    assert xmldoc.childNodes[-1].childNodes[0] is tab



# coding=utf-8
# Copyright (C) Duncan Macleod (2015)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Constants for `gwdetchar`
"""

import os
from collections import OrderedDict

from gwpy.segments import Segment

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

IFO = os.getenv('IFO', None)
ifo = os.getenv('ifo', IFO.lower() if IFO else None)
SITE = os.getenv('SITE', None)
site = os.getenv('site', SITE.lower() if SITE else None)

DEFAULT_SEGMENT_SERVER = os.getenv('DEFAULT_SEGMENT_SERVER',
                                   'https://segments.ligo.org')
S6_SEGMENT_SERVER = os.getenv('S6_SEGMENT_SERVER',
                              'https://segments-s6.ligo.org')

# -- Run epochs ---------------------------------------------------------------

EPOCH = OrderedDict([
    ('S6', Segment(931035615, 971622015)),
    ('ER7', Segment(1117400416, 1118329216)),
    ('ER8', Segment(1123858817, 1126623617)),
    ('O1', Segment(1126623617, 1136649617)),
    ('ER9', Segment(1152136817, 1152255617)),
    ('ER10', Segment(1161907217, 1164499217)),
    ('O2', Segment(1164499217, 1187733618)),
    ('O3', Segment(1187733618, 1366556418)),
    ('ER15', Segment(1366556418, 1368975618)),
    ('O4', Segment(1368975618, 2000000000)),
])


def latest_epoch():
    """Returns the name of the latest known epoch
    """
    epochs = sorted(EPOCH.items(), key=lambda x: x[1][0])
    return epochs[-1][0]


def gps_epoch(gpstime, default=None):
    """Returns the epoch name containing the given GPS time

    Parameters
    ----------
    gpstime : `float`, `int`
        the GPS time to match

    default : `str`, optional
        the epoch to return if ``gpstime`` doesn't match any
        known epoch, default is to raise `ValueError`

    Returns
    -------
    epoch : `str`
        the name of the containing epoch, or ``default`` if given

    Raises
    ------
    ValueError
        if ``gpstime`` doesn't match any known epoch, and ``default``
        was not given

    Examples
    --------
    >>> from gwdetchar.const import gps_epoch
    >>> print(gps_epoch(1126259462))
    'O1'
    >>> print(gps_epoch(0, default='ER0')
    'ER0'
    """
    for epoch in EPOCH:
        if float(gpstime) in EPOCH[epoch]:
            return epoch
    if default:
        return default
    raise ValueError("failed to match GPS time {0} to any "
                     "known epoch".format(gpstime))



# -*- coding: utf-8 -*-
# Copyright (C) Alex Urban (2019)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gwdetchar.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for `gwdetchar.io.datafind`
"""

import pytest

import numpy
from numpy import testing as nptest

from unittest import mock

from gwpy.timeseries import (TimeSeries, TimeSeriesDict)
from gwpy.segments import (Segment, DataQualityFlag)

from .. import datafind

__author__ = 'Alex Urban <alexander.urban@ligo.org>'


# global test objects

HOFT = TimeSeries(
    numpy.random.normal(loc=1, scale=.5, size=16384 * 66),
    sample_rate=16384, epoch=0, name='X1:TEST-STRAIN')

FLAG = DataQualityFlag(known=[(-33, 33)], active=[(-33, 33)],
                       name='X1:TEST-FLAG:1')


# -- make sure data can be read -----------------------------------------------

@mock.patch('gwpy.segments.DataQualityFlag.query', return_value=FLAG)
def test_check_flag(segserver):
    # attempt to query segments database for an analysis flag
    flag = 'X1:TEST-FLAG:1'
    assert datafind.check_flag(flag, gpstime=0, duration=64, pad=1) is True
    assert datafind.check_flag(flag, gpstime=800, duration=64, pad=1) is False


@mock.patch('gwpy.io.gwf.iter_channel_names', return_value=['X1:TEST-STRAIN'])
def test_remove_missing_channels(io_gwf):
    channels = datafind.remove_missing_channels(
        ['X1:TEST-STRAIN'], 'test.cache')
    assert channels == ['X1:TEST-STRAIN']

    with pytest.warns(UserWarning, match='is being removed because'):
        channels = datafind.remove_missing_channels(
            ['X1:TEST-STRAIN', 'X1:TEST-STRAIN_NONEXISTENT'], 'test.cache')
        assert channels == ['X1:TEST-STRAIN']


@mock.patch('gwpy.timeseries.TimeSeries.get', return_value=HOFT)
def test_get_data_from_NDS(tsget):
    # retrieve data
    start = 0
    end = 64
    channel = 'X1:TEST-STRAIN'
    data = datafind.get_data(channel, start, end)

    # test data products
    assert isinstance(data, TimeSeries)
    nptest.assert_array_equal(data.value, HOFT.value)


@mock.patch('gwpy.timeseries.TimeSeriesDict.get',
            return_value=TimeSeriesDict({'X1:TEST-STRAIN': HOFT}))
def test_get_data_dict_from_NDS(tsdget):
    # retrieve data
    start = 33
    end = 64
    channels = ['X1:TEST-STRAIN']
    data = datafind.get_data(channels, start, end)

    # test data products
    assert isinstance(data, TimeSeriesDict)
    nptest.assert_array_equal(data['X1:TEST-STRAIN'].value, HOFT.value)


@mock.patch('gwpy.io.datafind.find_urls')
@mock.patch('gwpy.timeseries.TimeSeries.read')
def test_get_data_from_cache(tsget, find_data):
    # set return values
    find_data.return_value = ['test.gwf']
    tsget.return_value = HOFT.crop(16, 48)

    # retrieve test frame
    start = 16
    end = start + 32
    channel = 'X1:TEST-STRAIN'
    frametype = 'X1_TEST'
    data = datafind.get_data(channel, start, end, frametype=frametype)

    # test data products
    assert isinstance(data, TimeSeries)
    assert data.duration.value == 32
    assert data.span == Segment(start, end)
    nptest.assert_array_equal(data.value, HOFT.crop(start, end).value)


@mock.patch('gwpy.io.datafind.find_urls')
@mock.patch('gwdetchar.io.datafind.remove_missing_channels')
@mock.patch('gwpy.timeseries.TimeSeriesDict.read')
def test_get_data_dict_from_cache(tsdget, remove, find_data):
    # set return values
    tsdget.return_value = TimeSeriesDict({
        'X1:TEST-STRAIN': HOFT.crop(16, 48)})
    remove.return_value = ['X1:TEST-STRAIN']
    find_data.return_value = ['test.gwf']
    # retrieve test frame
    start = 16
    end = start + 32
    channels = ['X1:TEST-STRAIN']
    data = datafind.get_data(channels, start, end, source=True)

    # test data products
    assert isinstance(data, TimeSeriesDict)
    assert data[channels[0]].duration.value == 32
    assert data[channels[0]].span == Segment(start, end)
    nptest.assert_array_equal(data[channels[0]].value,
                              HOFT.crop(start, end).value)


def test_get_data_bad_frametype():
    channel = 'X1:TEST-STRAIN'
    with pytest.raises(AttributeError) as exc:
        datafind.get_data(channel, start=0, end=32, frametype='bad_frametype')
    assert 'Could not determine observatory' in str(exc.value)



# coding=utf-8
# Copyright (C) LIGO Scientific Collaboration (2015-)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

import multiprocessing
import numpy
import os
import re
import sys

from scipy.stats import spearmanr

from gwpy.detector import ChannelList
from gwpy.io import nds2 as io_nds2

from .. import (cli, lasso as gwlasso)
from ..io import html as htmlio
from ..io.datafind import get_data

from sklearn import linear_model

from matplotlib import use
use('Agg')

# backend-dependent import
import matplotlib.pyplot as plt  # noqa: E402
from gwpy.plot import Plot  # noqa: E402

# initialize logger
PROG = ('python -m gwdetchar.lasso.old' if sys.argv[0].endswith('.py')
        else os.path.basename(sys.argv[0]))
LOGGER = cli.logger(name=PROG.split('python -m ').pop())


# -- parse command-line -------------------------------------------------------

def create_parser():
    """Create a command-line parser for this entry point
    """
    # initialize argument parser
    parser = cli.create_parser(
        prog=PROG,
        description=__doc__,
        formatter_class=cli.argparse.ArgumentDefaultsHelpFormatter,
    )
    psig = parser.add_argument_group('Signal processing options')

    # required arguments
    cli.add_gps_start_stop_arguments(parser)
    cli.add_ifo_option(parser)

    # optional arguments
    cli.add_nproc_option(parser, default=1)
    parser.add_argument(
        '-J',
        '--nproc-plot',
        type=int,
        default=None,
        help='number of processes to use for plotting',
    )
    parser.add_argument(
        '-o',
        '--output-dir',
        default=os.curdir,
        type=os.path.abspath,
        help='output directory for plots',
    )
    parser.add_argument(
        '-f',
        '--channel-file',
        type=os.path.abspath,
        help='path for channel file',
    )
    parser.add_argument(
        '-T',
        '--trend-type',
        default='minute',
        choices=['second', 'minute'],
        help='type of trend for correlation',
    )
    parser.add_argument(
        '-p',
        '--primary-channel',
        default='{ifo}:GDS-CALIB_STRAIN',
        help='name of primary channel to use',
    )
    parser.add_argument(
        '-P',
        '--primary-frametype',
        help='frametype for --primary-channel',
    )
    parser.add_argument(
        '-r',
        '--range-channel',
        default='{ifo}:DMT-SNSH_EFFECTIVE_RANGE_MPC.mean',
        help='name of range channel to use',
    )
    parser.add_argument(
        '-R',
        '--range-frametype',
        help='frametype for --range-channel',
    )
    parser.add_argument(
        '-O',
        '--remove-outliers',
        type=float,
        default=None,
        help='Std. dev. limit for removing outliers',
    )
    parser.add_argument(
        '-t',
        '--threshold',
        type=float,
        default=0.1,
        help='threshold for making a plot',
    )

    # signal processing arguments
    psig.add_argument(
        '-b',
        '--band-pass',
        type=float,
        nargs=2,
        default=None,
        metavar="FLOW FHIGH",
        help='lower and upper frequencies for bandpass on h(t)',
    )
    psig.add_argument(
        '-x',
        '--filter-padding',
        type=float,
        default=3.,
        help='amount of time (seconds) to pad data for filtering',
    )

    # return the argument parser
    return parser


# -- main code block ----------------------------------------------------------

def main(args=None):
    """Run the old lasso command-line interface
    """
    parser = create_parser()
    args = parser.parse_args(args=args)

    start = int(args.gpsstart)
    end = int(args.gpsend)
    pad = args.filter_padding
    try:
        flower, fupper = args.band_pass
    except TypeError:
        flower, fupper = None

    LOGGER.info('{} Slow Correlation {}-{}'.format(args.ifo, start, end))

    if args.primary_channel == '{ifo}:GDS-CALIB_STRAIN':
        args.primary_frametype = '%s_HOFT_C00' % args.ifo
    primary = args.primary_channel.format(ifo=args.ifo)
    rangechannel = args.range_channel.format(ifo=args.ifo)

    if not os.path.isdir(args.output_dir):
        os.makedirs(args.output_dir)
    os.chdir(args.output_dir)
    nprocplot = args.nproc_plot or args.nproc

    # load data
    LOGGER.info("-- Loading range data")
    rangets = get_data(
        rangechannel, start, end, frametype=args.range_frametype,
        verbose=True, nproc=args.nproc)

    if args.trend_type == 'minute':
        dstart, dend = rangets.span
    else:
        dstart = start
        dend = end

    LOGGER.info("-- Loading h(t) data")
    darmts = get_data(primary, dstart-pad, dend+pad, verbose=True,
                      frametype=args.primary_frametype, nproc=args.nproc)

    # get darm BLRMS
    LOGGER.debug("-- Filtering h(t) data")
    if args.trend_type == 'minute':
        stride = 60
    else:
        stride = 1
    if flower:
        darmblrms = (
            darmts.highpass(flower/2., fstop=flower/4.,
                            filtfilt=False, ftype='butter')
            .notch(60, filtfilt=False)
            .bandpass(flower, fupper, fstop=[flower/2., fupper*1.5],
                      filtfilt=False, ftype='butter')
            .crop(dstart, dend).rms(stride))
        darmblrms.name = '%s %s-%s Hz BLRMS' % (primary, flower, fupper)
    else:
        darmblrms = darmts.notch(60).crop(dstart, dend).rms(stride)
        darmblrms.name = '%s RMS' % primary

    if args.remove_outliers:
        LOGGER.debug(
            "-- Removing outliers above %f sigma" % args.remove_outliers)
        gwlasso.remove_outliers(darmblrms, args.remove_outliers)
        gwlasso.remove_outliers(rangets, args.remove_outliers)

    if args.trend_type == 'minute':
        # calculate the r value between the DARM BLRMS and the Range timeseries
        corr_p = numpy.corrcoef(rangets.value, darmblrms.value)[0, 1]
        # calculate the ρ value between the DARM BLRMS and the Range timeseries
        corr_s = spearmanr(rangets.value, darmblrms.value)[0]
    else:
        # for second trends, set correlation to 0 since sample rates differ
        corr_p = 0
        corr_s = 0

    # create scaled versions of data to compare to each other
    LOGGER.debug("-- Creating scaled data")
    rangescaled = rangets.detrend()
    rangerms = numpy.sqrt(sum(rangescaled**2.0)/len(rangescaled))
    darmscaled = darmblrms.detrend()
    darmrms = numpy.sqrt(sum(darmscaled**2.0)/len(darmscaled))

    # create scaled darm using the rms(range) and the rms(darm)
    if args.trend_type == 'minute':
        darmscaled *= (-rangerms / darmrms)

    # get aux data
    LOGGER.info("-- Loading auxiliary channel data")
    host, port = io_nds2.host_resolution_order(args.ifo)[0]
    if args.channel_file is None:
        channels = ChannelList.query_nds2('*.mean', host=host, port=port,
                                          type='m-trend')
    else:
        with open(args.channel_file, 'r') as f:
            channels = f.read().rstrip('\n').split('\n')
    nchan = len(channels)
    LOGGER.debug("Identified %d channels" % nchan)
    if args.trend_type == 'minute':
        frametype = '%s_M' % args.ifo  # for minute trends
    else:
        frametype = '%s_T' % args.ifo  # for second trends
    auxdata = get_data(
        list(map(str, channels)), dstart, dend, verbose=True,
        pad=0, frametype=frametype, nproc=args.nproc)

    gpsstub = '%d-%d' % (start, end-start)
    re_delim = re.compile('[:_-]')

    LOGGER.info("-- Processing channels")
    counter = multiprocessing.Value('i', 0)

    p1 = (.1, .1, .9, .95)
    p2 = (.1, .15, .9, .9)

    def process_channel(input_,):
        chan, ts = input_
        flat = ts.value.min() == ts.value.max()
        if flat:
            corr1 = None
            corr2 = None
            corr1s = None
            corr2s = None
            plot1 = None
            plot2 = None
            plot3 = None
        else:
            corr1 = numpy.corrcoef(ts.value, darmblrms.value)[0, 1]
            corr1s = spearmanr(ts.value, darmblrms.value)[0]
            if args.trend_type == 'minute':
                corr2 = numpy.corrcoef(ts.value, rangets.value)[0, 1]
                corr2s = spearmanr(ts.value, rangets.value)[0]
            else:
                corr2 = 0.0
                corr2s = 0.0
            # if all corralations are below threshold it does not plot
            if ((abs(corr1) < args.threshold)
               and (abs(corr1s) < args.threshold)
               and (abs(corr2) < args.threshold)
               and (abs(corr2s) < args.threshold)):
                plot1 = None
                plot2 = None
                plot3 = None
                return (chan, corr1, corr2, plot1,
                        plot2, plot3, corr1s, corr2s)

            plot = Plot(darmblrms, ts, rangets,
                        xscale="auto-gps", separate=True,
                        figsize=(12, 12))
            plot.subplots_adjust(*p1)
            plot.axes[0].set_ylabel('$h(t)$ BLRMS [strain]')
            plot.axes[1].set_ylabel('Channel units')
            plot.axes[2].set_ylabel('Sensitive range [Mpc]')
            for ax in plot.axes:
                ax.legend(loc='best')
                ax.set_xlim(start, end)
                ax.set_epoch(start)
            channelstub = re_delim.sub('_', str(chan)).replace('_', '-', 1)
            plot1 = '%s_TRENDS-%s.png' % (channelstub, gpsstub)
            try:
                plot.save(plot1)
            except (IOError, IndexError):
                plot.save(plot1)
            except RuntimeError as e:
                if 'latex' in str(e).lower():
                    plot.save(plot1)
                else:
                    raise
            plot.close()

            # plot auto-scaled verions
            tsscaled = ts.detrend()
            tsrms = numpy.sqrt(sum(tsscaled**2.0)/len(tsscaled))
            if args.trend_type == 'minute':
                tsscaled *= (rangerms / tsrms)
                if corr1 > 0:
                    tsscaled *= -1
            else:
                tsscaled *= (darmrms / tsrms)
                if corr1 < 0:
                    tsscaled *= -1
            plot = Plot(darmscaled, rangescaled, tsscaled,
                        xscale="auto-gps", figsize=[12, 6])
            plot.subplots_adjust(*p2)
            ax = plot.gca()
            ax.set_xlim(start, end)
            ax.set_epoch(start)
            ax.set_ylabel('Scaled amplitude [arbitrary units]')
            ax.legend(loc='best')
            plot2 = '%s_COMPARISON-%s.png' % (channelstub, gpsstub)
            try:
                plot.save(plot2)
            except (IOError, IndexError):
                plot.save(plot2)
            except RuntimeError as e:
                if 'latex' in str(e).lower():
                    plot.save(plot2)
                else:
                    raise
            plot.close()

            # plot scatter plots
            rangeColor = 'red'
            darmblrmsColor = 'blue'

            tsCopy = ts.reshape(-1, 1)
            rangetsCopy = rangets.reshape(-1, 1)
            darmblrmsCopy = darmblrms.reshape(-1, 1)

            darmblrmsReg = linear_model.LinearRegression()
            darmblrmsReg.fit(tsCopy, darmblrmsCopy)
            darmblrmsFit = darmblrmsReg.predict(tsCopy)

            rangeReg = linear_model.LinearRegression()
            rangeReg.fit(tsCopy, rangetsCopy)
            rangeFit = rangeReg.predict(tsCopy)

            fig = Plot(figsize=(12, 6))
            fig.subplots_adjust(*p2)
            ax = fig.add_subplot(121)
            ax.set_xlabel('Channel units')
            ax.set_ylabel('Sensitive range [Mpc]')
            yrange = abs(max(darmblrms.value) - min(darmblrms.value))
            upperLim = max(darmblrms.value) + .1 * yrange
            lowerLim = min(darmblrms.value) - .1 * yrange
            ax.set_ylim(lowerLim, upperLim)
            ax.text(.9, .1, 'r = ' + str('{0:.2}'.format(corr1)),
                    verticalalignment='bottom', horizontalalignment='right',
                    transform=ax.transAxes, color='black', size=20,
                    bbox=dict(boxstyle='square', facecolor='white', alpha=.75,
                              edgecolor='black'))
            fig.add_scatter(ts, darmblrms, color=darmblrmsColor)
            fig.add_line(ts, darmblrmsFit, color='black')

            ax = fig.add_subplot(122)
            ax.set_xlabel('Channel units')
            ax.set_ylabel('$h(t)$ BLRMS [strain]')
            ax.text(.9, .1, 'r = ' + str('{0:.2}'.format(corr2)),
                    verticalalignment='bottom', horizontalalignment='right',
                    transform=ax.transAxes, color='black', size=20,
                    bbox=dict(boxstyle='square', facecolor='white', alpha=.75,
                              edgecolor='black'))
            fig.add_scatter(ts, rangets, color=rangeColor)
            fig.add_line(ts, rangeFit, color='black')

            plot3 = '%s_SCATTER-%s.png' % (channelstub, gpsstub)
            try:
                fig.save(plot3)
            except (IOError, IndexError):
                fig.save(plot3)
            except RuntimeError as e:
                if 'latex' in str(e).lower():
                    fig.save(plot3)
                else:
                    raise
            plt.close(fig)

        # increment counter and print status
        with counter.get_lock():
            counter.value += 1
            pc = 100 * counter.value / nchan
            LOGGER.debug("Completed [%d/%d] %3d%% %-50s"
                         % (counter.value, nchan, pc, '(%s)' % str(chan)))
            sys.stdout.flush()
        return chan, corr1, corr2, plot1, plot2, plot3, corr1s, corr2s

    pool = multiprocessing.Pool(nprocplot)
    results = pool.map(process_channel, list(auxdata.items()))
    results.sort(key=lambda x: (x[1] is not None and max(abs(x[1]), abs(x[2]),
                 abs(x[6]), abs(x[7])) or 0, x[0]), reverse=True)

    with open('results.txt', 'w') as f:
        for ch, corr1, corr2, _, _, _, corr1s, corr2s in results:
            print('%s %s %s %s %s' % (
                ch, corr1, corr2, corr1s, corr2s), file=f)

    # -- write html
    trange = '%d-%d' % (start, end)
    title = '%s Slow Correlations: %s' % (args.ifo, trange)
    links = [trange] + [(s, '#%s' % s.lower())
                        for s in ['Parameters', 'Results']]
    (brand, class_) = htmlio.get_brand(args.ifo, 'Correlations', start)
    navbar = htmlio.navbar(links, class_=class_, brand=brand)
    page = htmlio.new_bootstrap_page(title=title, navbar=navbar)

    # header
    if flower:
        pstr = ('<code>%s</code> (band-limited %s-%s Hz)'
                % (primary, flower, fupper))
    else:
        pstr = primary
    if args.trend_type == 'minute':
        pstr += ' and <code>%s</code>' % rangechannel
    page.div(class_='pb-2 mt-3 mb-2 border-bottom')
    page.h1(title)
    page.p("This analysis searched %d channels for linear correlations with %s"
           % (nchan, pstr))
    page.div.close()

    # run parameters
    contents = [
        ('Primary channel',
         '{} ({})'.format(
             primary, args.primary_frametype.format(ifo=args.ifo))),
        ('Range channel',
         '{} ({})'.format(rangechannel, args.range_frametype or '-')),
        ('Band-pass', '{}-{}'.format(flower, fupper))]
    page.add(htmlio.parameter_table(contents, start=start, end=end))

    # results
    page.h2('Results', class_='mt-4', id_='results')
    r_blrms = "<i>r<sub>blrms</sub> </i>"
    r_range = "<i>r<sub>range</sub> </i>"
    r = "<i>r</i>"
    rho_blrms = "<i>&rho;<sub>blrms</sub> </i>"
    rho_range = "<i>&rho;<sub>range</sub> </i>"
    rho = "<i>&rho;</i>"
    Pearson_wikilink = htmlio.markup.oneliner.a(
        "Pearson's correlation coefficient",
        href="https://en.wikipedia.org/wiki/"
             "Pearson_product-moment_correlation_coefficient",
        rel="external")
    numpylink = htmlio.markup.oneliner.a(
        "<code>numpy.corrcoef</code>",
        href="http://docs.scipy.org/doc/numpy-1.10.1/reference/generated/"
             "numpy.corrcoef.html",
        rel="external")
    Spearman_wikilink = htmlio.markup.oneliner.a(
        "Spearman's correlation coefficient",
        href="https://en.wikipedia.org/wiki/"
             "Spearman%27s_rank_correlation_coefficient",
        rel="external")
    scipylink = htmlio.markup.oneliner.a(
        "<code>scipy.stats.spearmanr</code>",
        href="http://docs.scipy.org/doc/scipy-0.14.0/reference/generated/"
             "scipy.stats.spearmanr.html",
        rel="external")
    page.p("In the results below, all %s values are calculated as"
           " the square of %s using %s and all %s values are calculated"
           " as the square of %s using %s."
           % (r, Pearson_wikilink, numpylink, rho,
              Spearman_wikilink, scipylink))
    if args.trend_type == 'minute':
        page.p("%s and %s are reported for <code>%s</code>."
               " %s and %s are reported for <code>%s</code>."
               " The %s between these two channels is %.2f."
               " The %s between these two channels is %.2f."
               % (r_blrms, rho_blrms, primary, r_range, rho_range,
                  rangechannel, r, corr_p, rho, corr_s))

    page.div(id_='accordion')
    for i, (ch, corr1, corr2, plot1, plot2, plot3,
            corr1s, corr2s) in enumerate(results):
        if corr1 is None:
            h = '%s [flat]' % ch
        elif plot1 is None:
            h = ('%s [%s = %.2f, %s = %.2f] [%s = %.2f, %s = %.2f]'
                 ' [below threshold]'
                 % (ch, r_blrms, corr1, r_range, corr2, rho_blrms,
                    corr1s, rho_range, corr2s))
        elif args.trend_type == 'minute':
            h = ('%s [%s = %.2f, %s = %.2f] [%s = %.2f, %s = %.2f]'
                 % (ch, r_blrms, corr1, r_range, corr2, rho_blrms,
                    corr1s, rho_range, corr2s))
        else:
            h = '%s [%s = %.2f]' % (ch, r_blrms, corr1)
        if (corr1 is None) or (corr1 == 0) or (plot1 is None):
            context = 'bg-light'
        elif ((numpy.absolute(corr1) >= .6) or (numpy.absolute(corr1s) >= .6)
              or (numpy.absolute(corr2) >= .6)
              or (numpy.absolute(corr2s) >= .6)):
            context = 'text-white bg-danger'
        elif ((numpy.absolute(corr1) >= .4) or (numpy.absolute(corr1s) >= .4)
              or (numpy.absolute(corr2) >= .4)
              or (numpy.absolute(corr2s) >= .4)):
            context = 'text-white bg-warning'
        else:
            context = 'text-white bg-info'
        page.div(class_='card %s' % context)
        # heading
        page.div(class_='card-header')
        page.a(h, class_='collapsed card-link cis-link', href='#channel%d' % i,
               **{'data-toggle': 'collapse'})
        page.div.close()  # card-header
        # body
        page.div(id_='channel%d' % i, class_='collapse',
                 **{'data-parent': '#accordion'})
        page.div(class_='card-body')
        if corr1 is None:
            page.p("The amplitude data for this channel is flat"
                   " (does not change) for the chosen time period.")
        elif plot1 is None:
            page.p("Niether r nor rho are above the threshold of %.2f."
                   % (args.threshold))
        else:
            for p in (plot1, plot2, plot3):
                img = htmlio.FancyPlot(p)
                page.add(htmlio.fancybox_img(img))
        page.div.close()  # card-body
        page.div.close()  # collapse
        page.div.close()  # card
    page.div.close()  # accordion
    htmlio.close_page(page, 'index.html')  # save and close

    LOGGER.info("-- Process Completed")


# -- run from command-line ----------------------------------------------------

if __name__ == "__main__":
    main()
